from tequila import TequilaException, TequilaWarning

from tequila.circuit import QCircuit
from tequila.objective.objective import Variables
from tequila.quantumchemistry.qc_base import QuantumChemistryBase
from tequila.quantumchemistry.chemistry_tools import ClosedShellAmplitudes, Amplitudes
from tequila.quantumchemistry import ParametersQC, NBodyTensor
from tequila.quantumchemistry.chemistry_tools import ActiveSpaceData, OrbitalData

import copy
import numpy
import typing
import psi4
import warnings

from .fer_base import FermionicBase

from dataclasses import dataclass


class TequilaPsi4Exception(TequilaException):
    pass


class OpenVQEEPySCFException(TequilaException):
    pass


@dataclass
class Psi4Results:
    variables: dict = None  # psi4 variables dictionary, storing all computed values
    filename: str = None  # psi4 output file
    wfn: typing.Union[psi4.core.Wavefunction, psi4.core.CCWavefunction, psi4.core.CIWavefunction] = (
        None  # psi4 wavefunction
    )
    mol: psi4.core.Molecule = None


class QuantumChemistryPsi4(FermionicBase):
    def _make_psi4_active_space_data(self, active_orbitals, reference=None):
        """
        Small helper function
        Internal use only
        Parameters
        ----------
        active_orbitals: dictionary :
            dictionary with irreps as keys and a list of integers as values
            i.e. occ = {"A1":[0,1], "A2":[0]}
            means the occupied active space is made up of spatial orbitals
            0A1, 1A1 and 0A2
            as list: Give a list of spatial orbital indices
            i.e. occ = [0,1,3] means that spatial orbital 0, 1 and 3 are used
        reference: (Default value=None)
            List of orbitals which form the reference
            Can be given in the same format as active_orbitals
            If given as None then the first N_electron/2 orbitals are taken
            and the corresponding active orbitals are removed

        Returns
        -------
        Dataclass with active indices and reference indices (in spatial notation)

        """

        if active_orbitals is None:
            return None

        @dataclass
        class ActiveSpaceDataPsi4(ActiveSpaceData):
            frozen_docc: list = None  # frozen reference orbitals grouped by irrep (psi4 option, if None then the active space can not be represented by psi4)
            frozen_uocc: list = None  # frozen virtual orbtials grouped by irrep (psi4 option, if None then the active space can not be represented by psi4)

            def __str__(self):
                result = super().__str__()
                result += "{key:15} : {value:15} \n".format(key="frozen_docc", value=str(self.frozen_docc))
                result += "{key:15} : {value:15} \n".format(key="frozen_uocc", value=str(self.frozen_uocc))
                return result

            @property
            def psi4_representable(self):
                standard_ref = [self.reference_orbitals[0] == 0] + [
                    self.reference_orbitals[i] == self.reference_orbitals[i + 1] - 1
                    for i in range(len(self.reference_orbitals) - 1)
                ]
                return self.frozen_docc is not None and self.frozen_uocc is not None and all(standard_ref)

        # transform irrep notation to absolute ints
        active_idx = active_orbitals
        ref_idx = reference
        if isinstance(active_orbitals, dict):
            active_idx = []
            for key, value in active_orbitals.items():
                active_idx += [self.orbitals_by_irrep[key.upper()][i].idx_total for i in value]
        elif active_orbitals is None:
            active_idx = [i for i in range(self.n_orbitals)]

        standard_ref = sorted([i for i in range(self.n_electrons // 2)])
        if isinstance(reference, dict):
            ref_idx = []
            for key, value in reference.items():
                orbitals = [self.orbitals_by_irrep[key.upper()][i] for i in value]
                ref_idx += [x.idx_total for x in orbitals]
                ref_idx = sorted(ref_idx)
        elif reference is None:
            assert self.n_electrons % 2 == 0
            ref_idx = standard_ref

        # determine if the active space can be represented by psi4
        # reference needs to be the scf reference
        # all frozen orbitals need to be without gaps since we can not freeze individual orbitals
        frozen_docc = None
        if ref_idx == standard_ref:
            frozen_docc = [0] * self.nirrep
            frozen_docc_all = [self.orbitals[i] for i in standard_ref if self.orbitals[i].idx_total not in active_idx]
            for i, irrep in enumerate(self.irreps):
                sorted_array = sorted([x.idx_irrep for x in frozen_docc_all if x.irrep.upper() == irrep.upper()])
                frozen_docc[i] = len(sorted_array)
                if len(sorted_array) > 0 and (sorted_array[0] != 0 and sorted_array[-1] != len(sorted_array) - 1):
                    frozen_docc = None
                    print("active space not CAS type: frozen_docc of {} ".format(irrep), sorted_array)
                    break

        # and the same for the unoccupied ones
        frozen_uocc = None
        if frozen_docc is not None:
            frozen_uocc = [0] * self.nirrep
            frozen_uocc_all = [o for o in self.orbitals if o.idx_total not in active_idx + ref_idx]
            for i, irrep in enumerate(self.irreps):
                if irrep not in self.orbitals_by_irrep:
                    continue
                sorted_array = sorted([x.idx_irrep for x in frozen_uocc_all if x.irrep.upper() == irrep.upper()])
                frozen_uocc[i] = len(sorted_array)
                last = self.orbitals_by_irrep[irrep][-1]
                if len(sorted_array) > 0 and (
                    sorted_array[-1] != last.idx_irrep or sorted_array[-1] != sorted_array[0] + len(sorted_array) - 1
                ):
                    frozen_uocc = None
                    break

        return ActiveSpaceDataPsi4(
            active_orbitals=sorted(active_idx),
            reference_orbitals=sorted(ref_idx),
            frozen_docc=frozen_docc,
            frozen_uocc=frozen_uocc,
        )

    def __init__(
        self,
        parameters: ParametersQC,
        active_orbitals=None,
        reference_orbitals=None,
        frozen_orbitals=None,
        *args,
        **kwargs,
    ):
        """

        Parameters
        ----------
        parameters
        active_orbitals: dictionary :
            dictionary with irreps as keys and a list of integers as values
            i.e. occ = {"A1":[0,1], "A2":[0]}
            means the occupied active space is made up of spatial orbitals
            0A1, 1A1 and 0A2
            as list: Give a list of spatial orbital indices
            i.e. occ = [0,1,3] means that spatial orbital 0, 1 and 3 are used
        reference: (Default value=None)
            List of orbitals which form the reference
            Can be given in the same format as active_orbitals
            If given as None then the first N_electron/2 orbitals are taken
            and the corresponding active orbitals are removed
        args
        kwargs
        """
        psi4.core.clean()
        psi4.core.clean_options()
        psi4.core.clean_variables()

        psi4.core.clean()
        psi4.core.clean_options()
        psi4.core.clean_variables()

        self.psi4_mol = psi4.geometry(parameters.get_geometry_string())
        psi4.activate(self.psi4_mol)
        self._point_group = self.psi4_mol.point_group().symbol()
        if "point_group" in kwargs:
            self._point_group = kwargs["point_group"]

        self.energies = {}  # history to avoid recomputation
        self.logs = {}  # store full psi4 output

        # psi4 active space will be formed later
        super().__init__(
            parameters=parameters,
            active_orbitals=None,
            reference_orbitals=reference_orbitals,
            frozen_orbitals=[],
            *args,
            **kwargs,
        )

        oenergies = []
        for i in self.irreps:
            oenergies += [(i, j, x) for j, x in enumerate(self.orbital_energies(irrep=i))]

        oenergies = sorted(oenergies, key=lambda x: x[2])

        for x in self.orbitals:
            x.irrep = oenergies[x.idx_total][0]
            x.idx_irrep = oenergies[x.idx_total][1]
            x.energy = oenergies[x.idx_total][2]

        orbitals_by_irrep = {o.irrep: [] for o in self.orbitals}
        for o in self.orbitals:
            orbitals_by_irrep[o.irrep] += [o]

        self.orbitals_by_irrep = orbitals_by_irrep
        if reference_orbitals is None:
            reference_orbitals = [x.idx_total for x in self.reference_orbitals]

        if active_orbitals is None:
            active_orbitals = [i for i in range(self.n_orbitals)]

        if frozen_orbitals is None and self.parameters.frozen_core:
            frozen_orbitals = [i for i in range(self.parameters.get_number_of_core_electrons() // 2)]

        def list_to_irrep_dict(orbital_list):
            # assume we have been given a list of orbitals with their total indices instead of a dictionary with irreps
            active_dict = {}
            if orbital_list is None:
                return active_dict
            for x in orbital_list:
                orbital = self.orbitals[x]
                if orbital.irrep not in active_dict:
                    active_dict[orbital.irrep] = [orbital.idx_irrep]
                else:
                    active_dict[orbital.irrep] += [orbital.idx_irrep]
            return active_dict

        if not hasattr(active_orbitals, "keys"):
            active_orbitals = list_to_irrep_dict(active_orbitals)

        if not hasattr(reference_orbitals, "keys"):
            reference_orbitals = list_to_irrep_dict(reference_orbitals)

        if not hasattr(frozen_orbitals, "keys"):
            frozen_orbitals = list_to_irrep_dict(frozen_orbitals)

        # remove frozen-orbitals from active orbitals
        for k, v in frozen_orbitals.items():
            for x in v:
                if k in active_orbitals and x in active_orbitals[k]:
                    active_orbitals[k].remove(x)

        self.integral_manager.active_space = self._make_psi4_active_space_data(
            active_orbitals=active_orbitals, reference=reference_orbitals
        )
        # need to recompute
        # (psi4 won't take over active space information otherwise)
        self.compute_energy(method="hf", recompute=True, *args, **kwargs)
        self.ref_wfn = self.logs["hf"].wfn

        self.kwargs = kwargs

    @property
    def point_group(self):
        return self._point_group

    @property
    def nirrep(self):
        return self.ref_wfn.nirrep()

    def orbital_energies(self, irrep: typing.Union[int, str] = None, beta: bool = False, wfn=None):
        """
        Returns orbital energies of a given irrep
        or all orbital energies of all irreps (default)

        Parameters
        ----------
        irrep: int or str :
            int: specify the irrep by number (in cotton ordering)
            str: specify the irrep by name (like 'A1')
            specify from which irrep you want the orbital energies
            psi4 orders irreps in 'Cotton ordering'
            http://www.psicode.org/psi4manual/master/psithonmol.html#table-irrepordering
        beta: bool : (Default value=False)
            get the beta electrons

        Returns
        -------
        list or orbital energies
        """
        if wfn is None:
            wfn = self.ref_wfn

        if hasattr(irrep, "upper"):
            irrep = self.irreps.index(irrep.upper())

        if beta:
            tmp = psi4.driver.p4util.numpy_helper._to_array(wfn.epsilon_b(), dense=False)
        else:
            tmp = psi4.driver.p4util.numpy_helper._to_array(wfn.epsilon_a(), dense=False)

        if irrep is None or self.point_group.lower() == "c1":
            result = []
            for x in tmp:
                result += [x]
            return result
        else:
            return tmp[irrep]

    def initialize_integral_manager(self, *args, **kwargs):
        # get HF wavefunction and orbital coefficients
        if "ref_wfn" in kwargs:
            ref_wfn = kwargs["ref_wfn"]
            hf_energy = 0.0
        else:
            hf_energy = self.compute_energy(
                method="hf",
                ignore_active_space=True,
                options={"RHF__FAIL_ON_MAXITER": False},
                point_group=self.point_group,
            )
            ref_wfn = self.logs["hf"].wfn

        if ref_wfn.nirrep() != 1:
            wfn = ref_wfn.c1_deep_copy(ref_wfn.basisset())
        else:
            wfn = ref_wfn
        Ca = numpy.asarray(wfn.Ca())

        basisset = ref_wfn.basisset()
        mints = psi4.core.MintsHelper(basisset)

        # get integrals in atomic basis
        S = numpy.asarray(mints.ao_overlap())
        h = numpy.asarray(wfn.H())
        g = numpy.asarray(mints.ao_eri())
        c = float(wfn.variables()["NUCLEAR REPULSION ENERGY"])

        g = NBodyTensor(elems=numpy.asarray(g), ordering="chem")

        kwargs["overlap_integrals"] = S
        kwargs["two_body_integrals"] = g
        kwargs["one_body_integrals"] = h
        kwargs["constant_term"] = c
        kwargs["orbital_coefficients"] = Ca
        kwargs["basis_type"] = "GBS"
        kwargs["basis_name"] = self.parameters.basis_set

        irreps = [self.psi4_mol.point_group().char_table().gamma(i).symbol().upper() for i in range(ref_wfn.nirrep())]

        self.irreps = irreps

        oenergies = []
        for i in self.irreps:
            oenergies += [(i, j, x) for j, x in enumerate(self.orbital_energies(irrep=i, wfn=ref_wfn))]

        oenergies = sorted(oenergies, key=lambda x: x[2])

        if "orbitals" in kwargs:
            orbitals = kwargs["orbitals"]
        else:
            orbitals = None

        if orbitals is None:
            orbitals = [OrbitalData(idx_total=i) for i in range(h.shape[0])]

        for x in orbitals:
            x.irrep = oenergies[x.idx_total][0]
            x.idx_irrep = oenergies[x.idx_total][1]
            x.energy = oenergies[x.idx_total][2]

        kwargs["orbitals"] = orbitals

        # set some psi4 specific features
        self.ref_energy = hf_energy
        self.ref_wfn = self.logs["hf"].wfn
        self.irreps = irreps

        return super().initialize_integral_manager(*args, **kwargs)

    def compute_ccsd_amplitudes(self):
        return self.compute_amplitudes(method="ccsd")

    def _run_psi4(
        self,
        options: dict,
        method=None,
        return_wfn=True,
        point_group=None,
        filename: str = None,
        guess_wfn=None,
        ref_wfn=None,
        ignore_active_space=False,
        *args,
        **kwargs,
    ):
        psi4.core.clean()
        psi4.core.clean_variables()
        psi4.core.clean_options()

        if "threads" in kwargs:
            psi4.set_num_threads(nthread=kwargs["threads"])

        if filename is None:
            filename = "{}_{}.out".format(self.parameters.filename, method)

        psi4.core.set_output_file(filename)

        # easier guess read in
        if guess_wfn is not None:
            if isinstance(guess_wfn, QuantumChemistryPsi4):
                guess_wfn = guess_wfn.logs["hf"].wfn
            if isinstance(guess_wfn, str):
                guess_wfn = psi4.core.Wavefunction.from_file(guess_wfn)
            guess_wfn.to_file(guess_wfn.get_scratch_filename(180))  # this is necessary
            options["guess"] = "read"

        # prevent known flaws
        if "guess" in options and options["guess"].lower() == "read":
            options["basis_guess"] = False
            # additionally the outputfile needs to be the same
            # as in the previous guess
            # this can not be determined here
            # better pass down a guess_wfn

        mol = psi4.geometry(self.parameters.get_geometry_string())
        mol.set_multiplicity(self.parameters.multiplicity)
        if self.parameters.multiplicity != 1:
            raise TequilaPsi4Exception("Multiplicity != 1 no yet supported")
        mol.set_molecular_charge(self.parameters.charge)

        if point_group is not None:
            mol.reset_point_group(point_group.lower())

        if ref_wfn is None and hasattr(self, "ref_wfn"):
            ref_wfn = self.ref_wfn

        if point_group is not None and point_group.lower() == "c1":
            ref_wfn = None  # ref_wfn.c1_deep_copy(ref_wfn.basisset())  # CC will not converge otherwise
            guess_wfn = None

        psi4.activate(mol)

        psi4.set_options(options)
        if ref_wfn is None or method.lower() == "hf":
            energy, wfn = psi4.energy(name=method.lower(), return_wfn=return_wfn, molecule=mol)
        else:
            energy, wfn = psi4.energy(
                name=method.lower(), ref_wfn=ref_wfn, return_wfn=return_wfn, molecule=mol, guess_wfn=guess_wfn
            )
        self.energies[method.lower()] = energy
        self.logs[method.lower()] = Psi4Results(
            filename=filename, variables=copy.deepcopy(psi4.core.variables()), wfn=wfn, mol=mol
        )
        return energy, wfn

    def _extract_active_space(self, arr):
        """
        Internal function that helps to convert CCSD amplitudes if an actice space was set
        """
        if self.integral_manager.active_space is None:
            return arr
        elif (
            len(self.integral_manager.active_space.active_orbitals) == self.integral_manager.one_body_integrals.shape[0]
        ):
            return arr

        if isinstance(arr, ClosedShellAmplitudes):
            result = {}
            for k, v in arr.__dict__.items():
                if v is not None:
                    result[k] = self._extract_active_space(arr=v)

            return ClosedShellAmplitudes(**result)
        asd = self.integral_manager.active_space
        aocc = [i for i in asd.active_orbitals if i in asd.reference_orbitals]
        avir = [i for i in asd.active_orbitals if i not in asd.reference_orbitals]
        assert self.n_orbitals == len(aocc) + len(avir)
        assert self.n_electrons == len(aocc) * 2

        n_orb_total = len(self.orbitals)
        n_electrons_total = self.n_electrons + 2 * len(asd.frozen_reference_orbitals)
        nocc = n_electrons_total // 2
        nvirt = n_orb_total - nocc
        avir = [x - nocc for x in avir]
        nav = len(avir)
        nao = len(aocc)

        arr_shape = arr.shape
        final_shape = []
        active_sets = []

        for i in arr_shape:
            if i == nocc:
                final_shape.append(nao)
                active_sets.append(aocc)
            elif i == nvirt:
                final_shape.append(nav)
                active_sets.append(avir)
            else:
                assert i == nocc + nvirt
                final_shape.append(nao + nav)
                active_sets.append(aocc + avir)

        final_shape = tuple(final_shape)

        def func(*args):
            result = 1
            for i in range(len(args)):
                result *= args[i] in active_sets[i]
            return result

        c = numpy.fromfunction(function=numpy.vectorize(func), shape=arr_shape, dtype=numpy.int)

        return numpy.extract(condition=c, arr=arr).reshape(final_shape)

    # def compute_mp2_amplitudes(self, active_orbitals=None, *args, **kwargs) -> ClosedShellAmplitudes:
    #    return self._extract_active_space(super().compute_mp2_amplitudes(*args, **kwargs))

    def compute_amplitudes(
        self, method: str, options: dict = None, filename: str = None, *args, **kwargs
    ) -> typing.Union[Amplitudes, ClosedShellAmplitudes]:
        if options is None:
            options = {}

        options["basis"] = self.parameters.basis_set

        if method.lower() == "mp2":
            return self.compute_mp2_amplitudes()
        try:
            psi4.core.clean_options()
            psi4.core.clean_variables()
            energy, wfn = self._run_psi4(
                method=method,
                options=options,
                point_group="c1",
                ref_wfn=self.ref_wfn.c1_deep_copy(self.ref_wfn.basisset()),
                filename=filename,
                *args,
                **kwargs,
            )
            all_amplitudes = wfn.get_amplitudes()
            closed_shell = isinstance(wfn.reference_wavefunction(), psi4.core.RHF)
            if closed_shell:
                return self._extract_active_space(
                    ClosedShellAmplitudes(**{k: v.to_array() for k, v in all_amplitudes.items()})
                )
            else:
                assert self.integral_manager.trivial_active_space()  # only for closed-shell currently
                return Amplitudes(**{k: v.to_array() for k, v in all_amplitudes.items()})
        except Exception as err:
            raise TequilaPsi4Exception(
                "\nFailed to compute {} amplitudes.\n"
                "Make sure that you don't read in previous wavefunctions."
                "Active spaces might get you in trouble.".format(method)
            )

    def compute_energy(
        self, method: str = "fci", options=None, recompute: bool = True, ignore_active_space=False, *args, **kwargs
    ):
        if not recompute and method.lower() in self.energies and "point_group" not in kwargs:
            return self.energies[method.lower()]

        if options is None:
            options = {}

        options["basis"] = self.parameters.basis_set
        if (
            not ignore_active_space
            and not self.integral_manager.active_space_is_trivial()
            and self.active_space.psi4_representable
        ):
            options["frozen_docc"] = self.active_space.frozen_docc
            if sum(self.active_space.frozen_uocc) > 0 and method.lower() not in ["hf", "fci", "detci"]:
                print(
                    "There are known issues with some psi4 methods and frozen virtual orbitals. Proceed with fingers crossed for {}.".format(
                        method
                    )
                )
            options["frozen_uocc"] = self.active_space.frozen_uocc
        if (
            not ignore_active_space
            and not self.active_space.psi4_representable
            and not self.integral_manager.active_space_is_trivial()
        ):
            warnings.warn("Warning: Active space is not Psi4 representable", TequilaWarning)
        return self._run_psi4(method=method, options=options, ignore_active_space=ignore_active_space, *args, **kwargs)[
            0
        ]

    def __str__(self):
        result = super().__str__()
        result += "\nPsi4 Data\n"
        result += "{key:15} : {value:15} \n".format(
            key="Point Group (full)", value=self.psi4_mol.get_full_point_group().lower()
        )
        result += "{key:15} : {value:15} \n".format(key="Point Group (used)", value=self.point_group)
        result += "{key:15} : {value} \n".format(key="nirrep", value=self.nirrep)
        result += "{key:15} : {value} \n".format(key="irreps", value=self.irreps)
        result += "{key:15} : {value:15} \n".format(
            key="mos per irrep", value=str([len(self.orbital_energies(irrep=i)) for i in range(self.nirrep)])
        )

        return result

    @property
    def rdm1(self) -> tuple:
        return super().rdm1

    @property
    def rdm2(self) -> tuple:
        return super().rdm2

    def compute_rdms(
        self,
        U: QCircuit = None,
        variables: Variables = None,
        spin_free: bool = True,
        get_rdm1: bool = True,
        get_rdm2: bool = True,
        psi4_method: str = None,
        psi4_options: dict = {},
        *args,
        **kwargs,
    ):
        """
        Same functionality as qc_base.compute_rdms (look there for more information),
        plus the additional option to compute 1- and 2-RDM using psi4 by the keyword psi4_rdms

        Parameters
        ----------
        U :
             Quantum Circuit to achieve the desired state \\psi = U |0\\rangle, optional if psi4_rdms is set to True
        variables :
            If U is parametrized, then need to hand over a set of fixed variables
        spin_free :
            Set whether matrices should be spin-free (summation over spin) or defined by spin-orbitals
        get_rdm1, get_rdm2 :
            Set whether either one or both rdm1, rdm2 should be computed. If both are needed at some point,
            it is recommended to compute them at once.
            Note that whatever is specified in psi4_options has priority.
        psi4_method:
            Method to be run, currently only methods returning a CIWavefuntion are supported
            (e.g. "detci" + ex_level in options, or "fci", "cisdt", "casscf", but NOT "cisd")
        psi4_options:
           Options to be handed over to psi4, containing e.g. excitation level of "detci"-method.
           If "detci__opdm" for 1-RDM and "detci__tpdm" for 2-RDM are not included, the keywords get_rdm1, get_rdm2 are
           used (if both are specified, prioritizing psi4_options).

        Returns
        -------
        """
        if not psi4_method:
            return super().compute_rdms(
                U=U, variables=variables, spin_free=spin_free, get_rdm1=get_rdm1, get_rdm2=get_rdm2, *args, **kwargs
            )
        else:
            # Get 1- and 2-particle reduced density matrix via Psi4 CISD computation
            # If "cisd" is chosen, change to "detci" (default is excitation level 2 anyhow) to obtain a CIWavefunction
            if psi4_method.lower() == "cisd":
                print("Changed psi4_method from 'cisd' to 'detci' with ex_level=2 s.th. psi4 returns a CIWavefunction.")
                psi4_method = "detci"
            # Set options if not handed over
            psi4_options = {k.lower(): v for k, v in psi4_options.items()}  # set to lower-case for string comparison
            if "detci__opdm" not in psi4_options.keys():
                psi4_options.update({"detci__opdm": get_rdm1})
            if "detci__tpdm" not in psi4_options.keys():
                psi4_options.update({"detci__tpdm": get_rdm2})
            if psi4_method.lower() == "detci" and "detci__ex_level" not in psi4_options.keys():
                psi4_options.update({"detci__ex_level": 2})  # use CISD as default
                print(psi4_options)

            # Compute and set matrices
            self.compute_energy(psi4_method, options=psi4_options)
            wfn = self.logs[psi4_method].wfn
            if psi4_options["detci__opdm"]:
                rdm1 = psi4.driver.p4util.numpy_helper._to_array(wfn.get_opdm(-1, -1, "SUM", False), dense=True)
                self._rdm1 = rdm1
            if psi4_options["detci__tpdm"]:
                rdm2 = psi4.driver.p4util.numpy_helper._to_array(wfn.get_tpdm("SUM", False), dense=True)
                rdm2 = NBodyTensor(elems=rdm2, ordering="chem")
                rdm2.reorder(to="phys")  # RDMs in physics ordering (cp. to NBodyTensor in qc_base.py)
                rdm2 = 2 * rdm2.elems  # Factor 2 since psi4 normalizes 2-rdm by 1/2
                self._rdm2 = rdm2

    def perturbative_f12_correction(
        self,
        rdm1: numpy.ndarray = None,
        rdm2: numpy.ndarray = None,
        gamma: float = 1.4,
        n_ri: int = None,
        cabs_type: str = "active",
        cabs_options: dict = None,
        **kwargs,
    ) -> float:
        """
        Computes the spin-free [2]_R12 correction, needing only the 1- and 2-RDM of a reference method
        Requires either 1-RDM, 2-RDM or information to compute them in kwargs

        Parameters
        ----------
        rdm1 :
            1-electron reduced density matrix
        rdm2 :
            2-electron reduced density matrix
        gamma :
            f12-exponent, for a correlation factor f_12 = -1/gamma * exp[-gamma*r_12]
        n_ri :
            dimensionality of RI-basis; if None, then the maximum available via tensors / basis-set is used
        cabs_type :
            - either "active" for using a given basis set as is as approximative CBS (complete basis set), and specify
            OBS (orbital basis) by an active space
            - or "cabs+" for CABS+-approach as in
                Valeev, E. F. (2004). Improving on the resolution of the identity in linear R12 ab initio theories.
                Chemical Physics Letters, 395(4–6), 190–195. https://doi.org/10.1016/j.cplett.2004.07.061
                -> pass cabs_name in cabs_options
        cabs_options :
            dict, which needs at least {"cabs_name": some CABS basis set} if cabs_type=="cabs+"
        kwargs :
            e.g. RDM-information via {"U": QCircuit, "variables": optimal angles} if computation via VQE,
            or {"rdm__psi4_method": some CI method, "rdm__psi4_options": dict with psi4 options} if computation via
            psi4, compare to psi4_interface.compute_rdms
            one of the above needs to be passed if rdm1,rdm2 not yet computed

        Returns
        -------
            the f12 correction for the energy
        """
        from tequila.quantumchemistry.f12_corrections._f12_correction_psi4 import ExplicitCorrelationCorrectionPsi4

        correction = ExplicitCorrelationCorrectionPsi4(
            mol=self,
            rdm1=rdm1,
            rdm2=rdm2,
            gamma=gamma,
            n_ri=n_ri,
            cabs_type=cabs_type,
            cabs_options=cabs_options,
            **kwargs,
        )

        return correction.compute()



import abc
from os import path
from typing import Optional, List,Union,Tuple

import tequila as tq

from .color import RGB, DefaultColors, Color,Colors,ColorRange
from .state import CircuitState, CircuitStyle
from ... import graphical #import qpic_to_pdf,qpic_to_png
class Gate(abc.ABC):
    """
    Gate that can generate it's qpic visualization on its own as well as construct the matching tequila circuit.
    """

    @abc.abstractmethod
    def construct_circuit(self) -> tq.QCircuit:
        """Constructs the matching QCircuit"""
        pass

    @abc.abstractmethod
    def map_variables(self, variables) -> "Gate":
        pass


    def render(self, state: CircuitState, style: CircuitStyle) -> str:
        """
        This method renders the gate to a qpic string.
        This is the generic wrapper adding styling which is the same for each gate.
        When rendering subgates ALWAYS use this method.
        """
        out = self._render(state, style)
        if style.group_together:
            out += "\n" + state.all_names() + " TOUCH\n"
        return out


    @abc.abstractmethod
    def _render(self, state: CircuitState, style: CircuitStyle) -> str:
        """
        This method should return the qpic string for this gate, tailing \n is not needed!
        Do NOT call this method directly, instead use the generic render() method!
        """
        pass


    def export_qpic(self, filename: str, filepath: Optional[str] = None, style: dict = {}, labels: dict[int, str] = {},colors: dict[str, RGB] = {}, wire_colors: dict[int, Color] = {} , select:dict=None):
        """
        This method renders this gate/circuit into a complete qpic document and stores it in the given filepath (or cwd if omitted).
        Parameters:
            filename -- Name of the resulting file
            filepath -- If given, the qpic file will be stored in this directory.
            style -- Defines the styling that should be used to render the circuit.
            labels -- Dictionary mapping wire indices to names. These well be displayed instead of the index.
            colors -- Dictionary of additional colors that should be defined. The DefaultColors are prepended and can thus be overridden.
            wire_colors -- Dictionary mapping wire indices to a Color. If set the wire will be colored instead of defaulting to black.
        """
        result = ""
        style,colors = self._parse_CircuitStyle(style,colors)
        if not len(wire_colors) and select is not None:
            for s in select:
                if select[s] == 'B':
                    wire_colors[s] = "red"
        # add colors
        for color, rgb in DefaultColors.items():
            result += f"COLOR {color} {rgb.r} {rgb.g} {rgb.b}\n"

        for color, rgb in colors.items():
            result += f"COLOR {color} {rgb.r} {rgb.g} {rgb.b}\n"

        # add wires
        wires = self.used_wires()
        for wire in range(0, max(wires) + 1):
            name = "a" + str(wire)

            if wire in labels:
                label = labels[wire]
            else:
                label = str(wire)

            if wire in wire_colors:
                color = wire_colors[wire]
            else:
                color = "black"
            result += f"color={color} {name} W {label} \n"

        for wire in range(0, max(wires) + 1):
            result += f"a{wire} /\n"
        # add gates
        state = CircuitState(max(wires))
        result += self.render(state, style) + "\n"

        if filename is not None:
            extendedFilename = filename
            if not extendedFilename.endswith(".qpic"):
                extendedFilename = filename + ".qpic"
            if filepath is not None:
                extendedFilename = path.join(filepath, extendedFilename)
            with open(extendedFilename, "w") as file:
                file.write(result)

    def export_to(self,filename: str,**kwargs):
        """
        Shortcut to render this gate/circuit into a complete qpic/png/pdf document.
        """

        filename_tmp = filename.split(".")
        if 'filepath' in kwargs:
            filepath=kwargs['filepath']
            kwargs.pop('filepath')
        else: filepath = None
        if len(filename_tmp) == 1:
            ftype = "pdf"
            fname = filename
        else:
            ftype = filename_tmp[-1]
            fname = "".join(filename_tmp[:-1])
        if ftype == 'qpic':
            self.export_qpic(fname,**kwargs)
        elif ftype == 'pdf':
            self.export_qpic(fname,**kwargs)
            graphical.qpic_to_pdf(filename=fname,filepath=filepath)
        elif ftype == 'png':
            self.export_qpic(fname,**kwargs)
            graphical.qpic_to_png(filename=fname,filepath=filepath)
        else:
            raise tq.TequilaException(f'Extension {ftype} not supported directly. Try exporting to qpic and compiling to {ftype} yourself')
    @abc.abstractmethod
    def used_wires(self) -> List[int]:
        """
        This method should return all qubits that are used by this gate.
        """
        pass

    @abc.abstractmethod
    def dagger(self) -> "Gate":
        """
        This method should return the daggered gate.
        """
        pass

    def __str__(self):
        res = type(self).__name__
        U = self.used_wires()
        res += f': qubits {U}'
        return res
    
    def _parse_CircuitStyle(self,style:dict={},color_dict:dict={})->Tuple[CircuitStyle,dict]:
        def hex_to_three_letters(hex_color: str) -> str:
            """
            Deterministically converts a 6-digit hex color code to a 3-letter string.

            The string is generated by:
            1. Parsing the Red, Green, and Blue components.
            2. Using the value of each component to index into a custom character set.

            Args:
                hex_color: A string representing the hex color (e.g., '#FF5733' or 'FF5733').

            Returns:
                A 3-letter string derived from the hex color.
            """
            # 1. Clean and validate the input hex
            hex_color = hex_color.lstrip('#').upper()
            if len(hex_color) != 6:
                raise ValueError("Hex color must be 6 digits (e.g., FF5733).")

            # 2. Define the character pools (vowels/consonants/digits for variety)
            # The length of each pool must evenly divide 256 (the maximum R, G, or B value).
            # 256 / 8 = 32. This gives 32 possible indices for each pool.
            POOL_1 = "AEIOULMN"  # Vowels and common liquids/nasals
            POOL_2 = "BCDFGHKR"  # Common consonants
            POOL_3 = "PSTWXYZ1"  # Other consonants and a digit

            # Combine the pools for easy indexing (256/8 = 32 indices, so pool length should be 8)
            character_pools = [POOL_1, POOL_2, POOL_3]

            # 3. Split the hex into R, G, B components
            red_hex = hex_color[0:2]
            green_hex = hex_color[2:4]
            blue_hex = hex_color[4:6]

            hex_components = [red_hex, green_hex, blue_hex]
            
            result_letters = []

            # 4. Process each component
            for i in range(3):
                # Convert the 2-digit hex component to an integer (0-255)
                value = int(hex_components[i], 16)
                
                # Calculate the index using the modulo operator based on the pool size (8)
                # 0-255 mapped to 0-7
                index = value % len(character_pools[i])
                
                # Select the character
                letter = character_pools[i][index]
                result_letters.append(letter)

            return "".join(result_letters)
        def hex_to_rgb(hex:str)->Tuple:
            rgb = []
            for i in (0, 2, 4):
                decimal = int(hex[i:i+2], 16)/255
                rgb.append(decimal)
            return tuple(rgb)
        def parse_color(color:Union[Color,Tuple,List,str],color_dict:dict={})->Tuple[Color,dict]:
            if isinstance(color,Color):
                return color
            if isinstance(color,str):
                if len(color) and (len(color)==6 or color[0]=='#'): #it is a Hex RBG  
                    if color[0]=='#':
                        (r,g,b) = hex_to_rgb(color[1:])
                        name = hex_to_three_letters(color)
                    else:
                        (r,g,b) = hex_to_rgb(color)
                        name = hex_to_three_letters('#'+color)
                    color_dict[name]=RGB(r,g,b)
                    color = Color(name)
                else:
                    color = Color(color)
            elif isinstance(color,Union[List,Tuple]):
                assert len(color)==3, f"Three RGB components expected, received {len(color)}"
                color = RGB(color[0],color[1],color[2])
            else: raise Exception('No color format indified')
            return color,color_dict
        circut_style = CircuitStyle()
        if 'group_together' in style.keys():
            circut_style.group_together = style['group_together']
            style.pop('group_together')
        if 'parametrized' in style.keys():
            parametrized = style['parametrized']
            parametrized,color_dict = parse_color(parametrized,color_dict)
            if 'parametrized_text' in style:
                parametrized_text = style['parametrized_text']
                style.pop('parametrized_text')
            else: parametrized_text = 'black'
            parametrized_text,color_dict = parse_color(parametrized_text,color_dict)
            circut_style.parametrized_marking = Colors(parametrized_text,parametrized)
            style.pop('parametrized')
        if 'color_range' in style.keys():
            color_range = style['color_range']
            if isinstance(color_range,Union[List,Tuple]):
                if not len(color_range)==2:
                    raise Exception("Only 2 colors accepted")
                c0,color_dict = parse_color(color_range[0],color_dict)
                c1,color_dict = parse_color(color_range[1],color_dict)
                color_range = ColorRange(c0,c1)
            circut_style.color_range = color_range
            style.pop('color_range')
        if 'primary' in style.keys():
            primary = style['primary']
            primary,color_dict = parse_color(primary,color_dict)
            if 'primary_text' in style:
                primary_text = style['primary_text']
                style.pop('primary_text')
            else: primary_text = 'while'
            primary_text,color_dict = parse_color(primary_text,color_dict)
            circut_style.primary = Colors(primary_text,primary)
            style.pop('primary')
        if 'secondary' in style.keys():
            secondary = style['secondary']
            secondary,color_dict = parse_color(secondary,color_dict)
            if 'secondary_text' in style:
                secondary_text = style['secondary_text']
                style.pop('secondary_text')
            else: primary_text = 'while'
            secondary_text,color_dict = parse_color(secondary_text,color_dict)
            circut_style.secondary = Colors(secondary_text,secondary)
            style.pop('secondary')
        per = {}
        if 'personalize' in style.keys():
            personalize = style['personalize']
            style.pop('personalize')
            for gate_name,gate_color in personalize.items():
                gate_color,color_dict = parse_color(gate_color,color_dict)
                per[gate_name] = gate_color
        possibes = ['double','single','generic','UC','UR']
        for g in possibes:
            if g in style.keys():
                gate_color,color_dict = parse_color(style[g],color_dict)
                style.pop(g)
                if g+'_text' in style.keys():
                    text_color,color_dict = parse_color(style[g+'_text'],color_dict)
                else: text_color = Color('white')
                per[g] = Colors(text_color,gate_color)
        circut_style.personalized =  per
        return circut_style,color_dict


from tequila import TequilaException
from .HybridBase import HybridBase
from tequila.quantumchemistry import ParametersQC, NBodyTensor
import pyscf

import numpy, typing


class OpenVQEEPySCFException(TequilaException):
    pass



class PySCF(HybridBase):
    def __init__(self, parameters: ParametersQC,select: typing.Union[str,dict]={},
                 transformation: typing.Union[str, typing.Callable] = None,
                 *args, **kwargs):

        if "one_body_integrals" not in kwargs:

            geometry = parameters.get_geometry()
            pyscf_geomstring = ""
            for atom in geometry:
                pyscf_geomstring += "{} {} {} {};".format(atom[0], atom[1][0], atom[1][1], atom[1][2])

            if "point_group" in kwargs:
                point_group = kwargs["point_group"]
            else:
                point_group = None

            mol = pyscf.gto.Mole()
            mol.atom = pyscf_geomstring
            mol.basis = parameters.basis_set
            mol.charge = parameters.charge

            if point_group is not None:
                if point_group.lower() != "c1":
                    mol.symmetry = True
                    mol.symmetry_subgroup = point_group
                else:
                    mol.symmetry = False
            else:
                mol.symmetry = True

            mol.build(parse_arg=False)

            # solve restricted HF
            mf = pyscf.scf.RHF(mol)
            mf.verbose = False
            if "verbose" in kwargs:
                mf.verbose = kwargs["verbose"]

            mf.kernel()

            # only works if point_group is not C1
            # otherwise PySCF uses a different SCF object
            # irrep information is however not critical to tequila
            if hasattr(mf, "get_irrep_nelec"):
                self.irreps = mf.get_irrep_nelec()
            else:
                self.irreps = None

            orbital_energies = mf.mo_energy

            # compute mo integrals
            mo_coeff = mf.mo_coeff
            h_ao = mol.intor('int1e_kin') + mol.intor('int1e_nuc')
            g_ao = mol.intor('int2e', aosym='s1')
            S = mol.intor_symmetric("int1e_ovlp")
            g_ao = NBodyTensor(elems=g_ao, ordering="mulliken")

            self.pyscf_molecule = mol
            self.point_group = mol.symmetry_subgroup

            kwargs["overlap_integrals"] = S
            kwargs["two_body_integrals"] = g_ao
            kwargs["one_body_integrals"] = h_ao
            kwargs["orbital_coefficients"] = mo_coeff
            kwargs["orbital_type"] = "hf"

            if "nuclear_repulsion" not in kwargs:
                kwargs["nuclear_repulsion"] = mol.energy_nuc()
        super().__init__(parameters=parameters, transformation=transformation,select=select, *args, **kwargs)
    def compute_fci(self, *args, **kwargs):
        from pyscf import fci
        c, h1, h2 = self.get_integrals(ordering="chem")
        norb = self.n_orbitals
        nelec = self.n_electrons
        e, fcivec = fci.direct_spin1.kernel(h1, h2.elems, norb, nelec, **kwargs)
        return e + c

    def compute_energy(self, method: str, *args, **kwargs) -> float:
        method = method.lower()
        if 'hybrid' in method:
            return super().compute_energy(method,*args,**kwargs)
        elif method == "hf":
            return self._get_hf(do_not_solve=False, **kwargs).e_tot
        elif method == "mp2":
            return self._run_mp2(**kwargs).e_tot
        elif method == "cisd":
            hf = self._get_hf(do_not_solve=False, **kwargs)
            return self._run_cisd(hf=hf, **kwargs).e_tot
        elif method == "ccsd":
            return self._run_ccsd(**kwargs).e_tot
        elif method == "ccsd(t)":
            ccsd = self._run_ccsd(**kwargs)
            return ccsd.e_tot + self._compute_perturbative_triples_correction(ccsd=ccsd, **kwargs)
        elif method == "fci":
            return self.compute_fci(**kwargs)
        else:
            raise TequilaException("unknown method: {}".format(method))

    def _get_hf(self, do_not_solve=True, **kwargs):
        c, h1, h2 = self.get_integrals(ordering="mulliken")
        norb = self.n_orbitals
        nelec = self.n_electrons

        mo_coeff = numpy.eye(norb)
        mo_occ = numpy.zeros(norb)
        mo_occ[:nelec // 2] = 2

        pyscf_mol = pyscf.gto.M(verbose=0, parse_arg=False)
        pyscf_mol.nelectron = nelec
        pyscf_mol.incore_anyway = True  # ensure that custom integrals are used
        pyscf_mol.energy_nuc = lambda *args: c

        hf = pyscf.scf.RHF(pyscf_mol)
        hf.get_hcore = lambda *args: h1
        hf.get_ovlp = lambda *args: numpy.eye(norb)
        hf._eri = pyscf.ao2mo.restore(8, h2.elems, norb)

        if do_not_solve:
            hf.mo_coeff = mo_coeff
            hf.mo_occ = mo_occ
        else:
            hf.kernel(numpy.diag(mo_occ))

        return hf

    def _run_ccsd(self, hf=None, **kwargs):
        from pyscf import cc
        if hf is None:
            hf = self._get_hf()
        ccsd = cc.RCCSD(hf)
        ccsd.kernel()
        return ccsd

    def _compute_perturbative_triples_correction(self, ccsd=None, **kwargs) -> float:
        if ccsd is None:
            ccsd = self._run_ccsd(**kwargs)
        ecorr = ccsd.ccsd_t()
        return ecorr

    def _run_cisd(self, hf=None, **kwargs):
        from pyscf import ci
        if hf is None:
            hf = self._get_hf(**kwargs)
        cisd = ci.RCISD(hf)
        cisd.kernel()
        return cisd

    def _run_mp2(self, hf=None, **kwargs):
        from pyscf import mp
        if hf is None:
            hf = self._get_hf(**kwargs)
        mp2 = mp.MP2(hf)
        mp2.kernel()
        return mp2

    def __str__(self):
        base = super().__str__()
        try:
            if hasattr(self, "pyscf_molecule"):
                base += "{:15} : {} ({})\n".format("point_group", self.pyscf_molecule.groupname,
                                                   self.pyscf_molecule.topgroup)
            if hasattr(self, "irreps"):
                base += "{:15} : {}\n".format("irreps", self.irreps)
        except:
            return base
        return base


import numbers

from ..core import CircuitState, CircuitStyle
from ..quantum_chemistry import PairCorrelatorGate


class OrbitalRotatorGate(PairCorrelatorGate):
    """
        i,j correspond to Molecular Orbital index --> ((2*i,2*j),(2*i+1,2*j+1))
        only employed if n_qubits_is_double = False
    """

    def construct_circuit(self):
        return self._dummy.UR(self.i, self.j, control=self.control, assume_real=self.assume_real, angle=self.angle)


    def _render(self, state: CircuitState, style: CircuitStyle) -> str:
        # styling
        shape = 2
        tcol = style.primary.text_color
        gcol = style.primary.gate_color

        # angle
        gcol, tcol = self.render_angle(style, gcol, tcol)
        if 'UR' in style.personalized:
            gcol = style.personalized['UR'].gate_color
            tcol = style.personalized['UR'].text_color
        # TODO: param is not used again in the original??
        param = self.angle
        if isinstance(param, numbers.Number):
            param = "{:1.2f}".format(param)

        # rendering
        result = " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i, shape=shape,gcol=gcol,tcol="{" + tcol.name + "}",op="")
        result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j, shape=shape,gcol=gcol,tcol="{" + tcol.name + "}",op="")

        return result



from typing import List,Union
from copy import deepcopy
from ..core import Gate
from sunrise.graphical import generic
from ..quantum_chemistry import PairCorrelatorGate
from ..quantum_chemistry import SingleExcitation
from ..quantum_chemistry import GenericExcitation
from ..generic.gate import GenericGate
from ..quantum_chemistry import DoubleExcitation
from ..quantum_chemistry import OrbitalRotatorGate
from ...fermionic_operations.circuit import FCircuit
from tequila import QCircuit
from tequila.circuit.gates import X
from typing import Union

class GraphicalCircuit(Gate):
    """
    Basic wrapper class for multiple gates.
    Gates can be added in a similar as with tequila's QCircuit: Circuit += gate
    """

    gates: List[Gate]

    def __init__(self, gates: List[Gate] = []):
        if gates is not None: self.gates = gates

    def __add__(self, other: Gate):
        self.gates.append(other)
        return self
    def __str__(self):
        res = "GraphicalCircuit:\n"
        for gate in self.gates:
            res += gate.__str__() +'\n'
        return res
    def dagger(self) -> "GraphicalCircuit":
        return GraphicalCircuit([gate.dagger() for gate in reversed(self.gates)])

    def _render(self, state, style) -> str:
        output = ""
        sim = self._simplify_GraphicalCircuit()
        for gate in sim.gates:
            output += gate.render(state, style) + " \n"
        return output

    def used_wires(self) -> List[int]:
        wires = set()
        for gate in self.gates:
            wires.update(gate.used_wires())
        return list(wires)

    def construct_circuit(self) -> QCircuit:
        U = QCircuit()
        for gate in self.gates:
            U += gate.construct_circuit()
        return U

    def map_variables(self, variables) -> "GraphicalCircuit":
        gates = []
        for gate in self.gates:
            gates.append(gate.map_variables(variables))
        return GraphicalCircuit(gates)
    def _simplify_GraphicalCircuit(self) -> "GraphicalCircuit":
        cp = deepcopy(self)
        for g in range(len(self.gates)-1)[::-1]:
            if isinstance(self.gates[g],generic.GenericGate) and isinstance(self.gates[g+1],generic.GenericGate):
                cp.gates[g].U += cp.gates[g+1].U
                cp.gates[g].qubits = [*set(cp.gates[g].qubits+cp.gates[g+1].qubits)]
                cp.gates.pop(g+1)
        return cp
        
    @classmethod
    def from_circuit(cls,U:Union[QCircuit,FCircuit,'GraphicalCircuit'], n_qubits_is_double: bool = False, *args, **kwargs) -> 'GraphicalCircuit':
        if isinstance(U,QCircuit):
            return cls.from_qcircuit(U=U,n_qubits_is_double=n_qubits_is_double,*args,**kwargs)
        elif isinstance(U,FCircuit):
            return cls.from_fcircuit(U=U,n_qubits_is_double=n_qubits_is_double,*args,**kwargs)
        elif isinstance(U,'GraphicalCircuit'):
            return cls(gates=U.gates)
        else: raise Exception(f'Type {type(U).__name__} not recognized, expected tq QCircuit or sun FCircuit')
            

    @classmethod
    def from_qcircuit(cls,U:QCircuit, n_qubits_is_double: bool = False, *args, **kwargs) -> 'GraphicalCircuit':
        """
        Converts a tequila QCircuit into a GraphicalCircuit object, which is renderable with `export_qpic`.
        """
        circuit = U._gates
        res = []
        was_ur = False
        for i, gate in enumerate(circuit):
            if gate._name == 'FermionicExcitation':
                index = ()
                for pair in gate.indices:
                    for so in pair:
                        index += (so,)
                if len(index) == 2:
                    if was_ur:
                        was_ur = False
                        continue
                    elif gate != circuit[-1] and circuit[i + 1]._name == 'FermionicExcitation' and len(circuit[i + 1].indices[0]) == 2 and circuit[i + 1].indices[0][0] // 2 == index[0] // 2 and circuit[i + 1].indices[0][1] // 2 == index[1] // 2 and not n_qubits_is_double:  # hope you enjoy this conditional
                        res.append(OrbitalRotatorGate(index[0] // 2, index[1] // 2, angle=gate._parameter, *args, **kwargs))
                        was_ur = True
                    else:
                        res.append(SingleExcitation(index[0], index[1], angle=gate._parameter,n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
                elif len(index) ==4:
                    if index[0] // 2 == index[2] // 2 and index[1] // 2 == index[3] // 2 and not n_qubits_is_double:  ## TODO: Maybe generalized for further excitations
                        res.append(PairCorrelatorGate(index[0] // 2, index[1] // 2, angle=gate._parameter, *args, **kwargs))
                    else:
                        res.append(DoubleExcitation(index[0], index[1], index[2], index[3], angle=gate._parameter, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
                else: res.append(GenericExcitation(indices=index,angle=gate._parameter, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
            elif gate._name == 'QubitExcitation':
                index = gate._target
                if len(index) == 2:
                    if was_ur:
                        was_ur = False
                        continue
                    elif gate != circuit[-1] and circuit[i + 1]._name == 'QubitExcitation' and len(circuit[i + 1]._target) == 2 and circuit[i + 1]._target[0] // 2 == index[0] // 2 and circuit[i + 1]._target[1] // 2 == index[1] // 2 and not n_qubits_is_double:  # hope you enjoy this conditional
                        res.append(OrbitalRotatorGate(index[0] // 2, index[1] // 2, angle=gate._parameter, *args, **kwargs))
                        was_ur = False
                    else:
                        res.append(SingleExcitation(index[0], index[1], angle=gate._parameter,n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
                else:
                    if index[0] // 2 == index[2] // 2 and index[1] // 2 == index[3] // 2 and not n_qubits_is_double:  ## TODO: Maybe generalized for further excitations
                        res.append(PairCorrelatorGate(index[0] // 2, index[1] // 2, angle=gate._parameter,*args, **kwargs))
                    else:res.append(DoubleExcitation(index[0], index[1], index[2], index[3], angle=gate._parameter, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
            else:
                res.append(GenericGate(U=QCircuit(gates=[gate]), name="simple", n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
        return cls(gates=res) 
    
    @classmethod
    def from_fcircuit(cls,U:FCircuit, n_qubits_is_double: bool = False, *args, **kwargs) -> 'GraphicalCircuit':
        """
        Converts a tequila FCircuit into a GraphicalCircuit object, which is renderable with `export_qpic`.
        """

        circuit = U._gates
        res = []
        if U.initial_state is not None:
            E = X(target=U.qubits)
            # E.qubits = [*range(U.initial_state.n_qubits)]
            res.append(GenericGate(U=E, name="initialstate", n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
        for gate in circuit:
            index = ()
            for pair in gate.indices:
                for so in pair:
                    index += (so,)
            if gate._name == 'FermionicExcitation':
                if len(index) == 1:
                    res.append(SingleExcitation(index[0][0], index[0][1], angle=gate.variables,n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
                elif len(index) == 2:
                    res.append(DoubleExcitation(index[0][0], index[0][1], index[1][0], index[1][1], angle=gate.variables, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
                else: res.append(GenericExcitation(indices=index,angle=gate.variables, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
            elif gate._name == 'UR':
                if not n_qubits_is_double:
                    res.append(OrbitalRotatorGate(index[0][0] // 2, index[0][1] // 2, angle=gate.variables, *args, **kwargs))
                else: 
                    res.append(SingleExcitation(index[0][0], index[0][1], angle=gate.variables,n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
                    res.append(SingleExcitation(index[1][0], index[1][1], angle=gate.variables,n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
            elif gate._name == 'UC':
                if not n_qubits_is_double:
                    res.append(PairCorrelatorGate(index[0][0] // 2, index[0][1] // 2, angle=gate.variables,*args, **kwargs))
                else: 
                    res.append(DoubleExcitation(index[0][0], index[0][1], index[0][0], index[1], angle=gate.variables, n_qubits_is_double=n_qubits_is_double, *args,**kwargs))
            else:
                res.append(GenericGate(U=gate, name="simple", n_qubits_is_double=n_qubits_is_double, *args, **kwargs))
        return cls(gates=res)


import tequila as tq
import sunrise as sun
import pytest
import numpy as np
import openfermion as of
import scipy

# @pytest.mark.skipif(condition=not HAS_PSI4 and not HAS_PYSCF, reason="psi4/pyscf not found")
@pytest.mark.parametrize("backend",tq.chemistry.INSTALLED_QCHEMISTRY_BACKENDS)
def test_hcb_measurement_linearH4_scenario2(backend):
    if backend == 'base':
        pytest.skip("Base")
    # Create the molecule
    mol = tq.Molecule(geometry="h 0.0 0.0 0.0\nh 0.0 0.0 1.5\nh 0.0 0.0 3.0\nh 0.0 0.0 4.5", basis_set="sto-3g", backend=backend).use_native_orbitals()
    H = mol.make_hamiltonian()

    # Create circuit
    U0 = mol.make_ansatz(name="SPA", edges=[(0,1),(2,3)])
    UR1 = mol.UR(0,1,angle=np.pi/2) + mol.UR(2,3,angle=np.pi/2) + mol.UR(0,3,angle=-0.2334) + mol.UR(1,2,angle=-0.2334)
    energy1 = -1.979853695572642
    variables1 = {((0, 1), 'D', None): -0.6621692220931846, ((2, 3), 'D', None): -0.6621692220931825}

    UR2 = mol.UR(1,2,angle=np.pi/2) + mol.UR(0,3,angle=np.pi/2)
    UR2+= mol.UR(0,1,angle="x") + mol.UR(0,2,angle="y") + mol.UR(1,3,angle="xx") + mol.UR(2,3,angle="yy") + mol.UR(1,2,angle="z") + mol.UR(0,3,angle="zz")
    UC2 = mol.UC(1,2,angle="b") + mol.UC(0,3,angle="c")
    U = U0 + UR1.dagger() + UR2 + UC2 + UR2.dagger()
    variables = {((0, 1), 'D', None): -0.644359150621798, ((2, 3), 'D', None): -0.644359150621816, "x": 0.4322931478168998, "y": 4.980327764918099e-14,
                 "xx": -3.07202211218271e-14, "yy": 0.7167447375727501, "z": -3.982666230146327e-14, "zz": 1.2737831353027637e-13, "c": -0.011081251246998072,
                 "b": 0.49719805420976604}
    E = tq.ExpectationValue(H=H, U=U)
    energy = tq.simulate(E, variables=variables)

    # Create rotators
    graphs = [
        [(0,1),(2,3)],
        [(0,3),(1,2)],
        [(0,2),(1,3)]
    ]
    rotators = []
    for graph in graphs:
        UR = tq.QCircuit()
        for edge in graph:
            UR += mol.UR(edge[0], edge[1], angle=np.pi/2)
        rotators.append(UR)

    # Test rotation of H and U simultaneously
    HX = sun.fold_rotators(mol, rotators[0]).make_hamiltonian()
    EX = tq.ExpectationValue(H=HX, U=U+rotators[0])
    assert np.isclose(energy, tq.simulate(EX, variables))

    # Test measurement of H rotated in SPA with SPA circuit
    E_test = tq.ExpectationValue(U=U0+rotators[0].dagger(), H=mol.make_hamiltonian())
    tmol = sun.fold_rotators(mol, rotators[0])
    hcb_mol, res_mol = sun.get_hcb_part(tmol)
    EX = tq.ExpectationValue(U=U0+rotators[0].dagger() + rotators[0], H=hcb_mol.make_hamiltonian())
    assert np.isclose(tq.simulate(EX, variables1), tq.simulate(E_test, variables1))

    # Apply the measurement protocol
    result = sun.rotate_and_hcb(molecule=mol, circuit=U, variables=variables, rotators=rotators, target=energy, silent=True)

    test_energy = 0
    for i,molecule in enumerate(result[0]):
        E = tq.ExpectationValue(H=molecule.make_hamiltonian(), U=U+rotators[i])
        test_energy += tq.simulate(E, variables=variables)
    assert np.isclose(test_energy, energy, 10**-3)

@pytest.mark.parametrize("backend",tq.chemistry.INSTALLED_QCHEMISTRY_BACKENDS)
def test_hcb_measurement_linearH4_scenario1(backend):
    if backend == 'base':
        pytest.skip("Base")
    # Create the molecule
    mol = tq.Molecule(geometry="h 0.0 0.0 0.0\nh 0.0 0.0 1.5\nh 0.0 0.0 3.0\nh 0.0 0.0 4.5", basis_set="sto-3g", backend=backend).use_native_orbitals()
    fci = mol.compute_energy("fci")
    H = mol.make_hamiltonian()

    # Create true wave function
    Hof = H.to_openfermion()
    Hsparse = of.linalg.get_sparse_operator(Hof)
    v,vv = scipy.sparse.linalg.eigsh(Hsparse, sigma=fci)
    wfn = tq.QubitWaveFunction.from_array(vv[:,0])
    energy = wfn.inner(H * wfn).real

    # Create rotators
    graphs = [
        [(0,1),(2,3)],
        [(0,3),(1,2)],
        [(0,2),(1,3)]
    ]
    rotators = []
    for graph in graphs:
        UR = tq.QCircuit()
        for edge in graph:
            UR += mol.UR(edge[0], edge[1], angle=np.pi/2)
        rotators.append(UR)

    # Test rotation of H and U simultaneously
    HX = sun.fold_rotators(mol, rotators[0]).make_hamiltonian()
    EX = tq.ExpectationValue(H=HX, U=rotators[0])
    assert np.isclose(energy, tq.simulate(EX, initial_state=wfn))

    # Apply the measurement protocol
    result = sun.rotate_and_hcb(molecule=mol, rotators=rotators, target=fci, initial_state=wfn, silent=True)

    test_energy = 0
    for i,molecule in enumerate(result[0]):
        E = tq.ExpectationValue(H=molecule.make_hamiltonian(), U=rotators[i])
        test_energy += tq.simulate(E, initial_state=wfn)
    assert np.isclose(test_energy, energy, 10**-3)


import tequila as tq
import pytest
import numpy
import sunrise as sn
@pytest.mark.parametrize("system",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize("select",["FBFBFFBFBFBFB","BBFFBBFFBBFF"])
@pytest.mark.parametrize("two_qubit",[True,False])
def test_hamiltonian(system,select,two_qubit):
    mol= sn.Molecule(geometry=system,basis_set="sto-6g",select=select,backend='pyscf',two_qubit=two_qubit,nature='hybrid')
    tqmol=tq.Molecule(basis_set="sto-6g",geometry=system,backend='pyscf')
    edges = mol.get_spa_edges()
    U1=tqmol.make_ansatz("SPA",edges=edges)
    H1 = tqmol.make_hamiltonian()
    U = mol.make_ansatz("SPA",edges=edges)
    H = mol.make_hamiltonian()
    E = tq.ExpectationValue(H=H, U=U)
    E1 = tq.ExpectationValue(H=H1, U=U1)
    result = tq.minimize(E,silent=True)
    result1 = tq.minimize(E1,silent=True)
    assert numpy.isclose(result1.energy,result.energy,10**-5)

@pytest.mark.parametrize("two_qubit",[True,False])
@pytest.mark.parametrize("transformation",["Jordan-Wigner","reordered-Jordan-Wigner"])
def test_opt_SPA(two_qubit,transformation):
    mol= sn.Molecule(geometry="H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8",basis_set="sto-6g",select="BBFFBBFFBBFF",backend='pyscf',two_qubit=two_qubit,transformation=transformation,nature='hybrid') #could be any select
    tqmol=tq.Molecule(basis_set="sto-6g",geometry="H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8",backend='pyscf',transformation=transformation)
    edges = mol.get_spa_edges()
    initial_guess = mol.get_spa_guess().T
    tqopt = tq.quantumchemistry.optimize_orbitals(molecule=tqmol,circuit=tqmol.make_ansatz("HCB-SPA",edges=edges),silent=True,initial_guess=initial_guess,use_hcb=True)
    opt = sn.optimize_orbitals(molecule=mol,circuit=mol.make_ansatz("SPA",edges=edges),silent=True,initial_guess=initial_guess)
    assert numpy.isclose(tqopt.energy,opt.energy,10**-5)
@pytest.mark.parametrize("system",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize("select",["FFFFFFFFFFF","FBFBFBFBFB"])
@pytest.mark.parametrize("excit",[ [(0,4)] , [(2,4),(3,5)], [(2,6),(3,7),(1,5)] , [(2,4),(3,5),(1,7),(0,6)] ])
def test_exc_gate(system,select,excit):
    mol_mix = sn.Molecule(geometry=system, basis_set="sto-6g", select=select,condense=False,backend='pyscf',nature='hybrid')
    mol_jw = tq.Molecule(basis_set="sto-6g", geometry=system,backend='pyscf')

    U_jw = mol_jw.prepare_reference()
    U_mix = mol_mix.prepare_reference()

    U_jw += mol_jw.make_excitation_gate(indices=excit, angle="a")
    U_mix += mol_mix.make_excitation_gate(indices=excit, angle="a")

    H_jw = mol_jw.make_hamiltonian()
    H_mix = mol_mix.make_hamiltonian()

    E_jw = tq.ExpectationValue(H=H_jw, U=U_jw)
    E_mix = tq.ExpectationValue(H=H_mix, U=U_mix)

    result_jw = tq.minimize(E_jw)
    result_mix = tq.minimize(E_mix)

    U_mix += mol_mix.transformation.hcb_to_me(bos=True)
    U_mix.n_qubits=U_jw.n_qubits
    wfn_jw = tq.simulate(U_jw,variables=result_jw.variables)
    wfn_mix = tq.simulate(U_mix,variables=result_mix.variables)
    F = abs(wfn_jw.inner(wfn_mix))

    assert numpy.isclose(F, 1.0, 10 ** -4)
    assert numpy.isclose(result_mix.energy, result_jw.energy, 10 ** -4)
@pytest.mark.parametrize("system",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize("hcb_optimization",[True, False])
@pytest.mark.parametrize("order",[1, 2])
def test_UpCCGD_BOS(system,hcb_optimization,order):
    mol = sn.Molecule(geometry=system,basis_set="sto-6g",select="",condense=False,backend="pyscf",nature='hybrid')
    tqmol = tq.Molecule(geometry=system,basis_set="sto-6g",backend="pyscf")

    H = mol.make_hamiltonian()
    tqH = tqmol.make_hamiltonian()

    U = mol.make_ansatz("UPCCGD",include_reference=True,hcb_optimization=hcb_optimization,order=order)
    tqU = tqmol.make_ansatz("UPCCGD",include_reference=True,hcb_optimization=hcb_optimization,order=order)

    E = tq.ExpectationValue(H=H, U=U)
    tqE = tq.ExpectationValue(H=tqH, U=tqU)
    res = tq.minimize(E, silent=True)
    tqres = tq.minimize(tqE, silent=True)
    wfn =tq.simulate(U + mol.hcb_to_me(bos=True), variables=res.variables)
    tqwfn = tq.simulate(tqU , variables=tqres.variables)
    assert numpy.isclose(res.energy,tqres.energy,atol=1.e-4)
    assert numpy.isclose(abs(tqwfn.inner(wfn)),1.0,atol=1.e-4)
@pytest.mark.parametrize("system",["H 0.0 0.0 0.0\nH 0.0 0.0 1.6\nH 0.0 0.0 3.2\nH 0.0 0.0 4.8","H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2"])
@pytest.mark.parametrize("hcb_optimization",[True, False])
def test_UpCCGSD_FER(system,hcb_optimization):
    mol = sn.Molecule(geometry=system, basis_set="sto-6g", select="FFFFFFFFFFFFFFFFFFFFF", condense=False,backend='pyscf',nature='hybrid')
    tqmol = tq.Molecule(geometry=system, basis_set="sto-6g",backend='pyscf')

    H = mol.make_hamiltonian()
    tqH = tqmol.make_hamiltonian()

    U = mol.make_ansatz("UPCCGSD", include_reference=True, hcb_optimization=hcb_optimization)
    tqU = tqmol.make_ansatz("UPCCSGD", include_reference=True, hcb_optimization=hcb_optimization)

    E = tq.ExpectationValue(H=H, U=U)
    tqE = tq.ExpectationValue(H=tqH, U=tqU)
    res = tq.minimize(E, silent=True)
    tqres = tq.minimize(tqE, silent=True)
    wfn = tq.simulate(U , variables=res.variables)
    tqwfn = tq.simulate(tqU, variables=tqres.variables)
    assert numpy.isclose(res.energy, tqres.energy, atol=1.e-4)
    assert numpy.isclose(abs(tqwfn.inner(wfn)), 1.0, atol=1.e-4)
#def test_UCCSD_FER():
#    '''
#    Take care, expensive test
#    '''
#    mol = sn.Molecule(geometry="H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2", basis_set="sto-6g", select="FFFFFFFFFFFFFFFFFFFFF", condense=False,backend='pyscf',nature='hybrid')
#    tqmol = tq.Molecule(geometry="H 0. 0. 0.\n Be 0. 0. 1.6\n H 0. 0. 3.2", basis_set="sto-6g",backend='pyscf')
#
#    H = mol.make_hamiltonian()
#    tqH = tqmol.make_hamiltonian()
#
#    U = mol.make_ansatz("UCCSD",add_singles=False)
#    tqU = tqmol.make_ansatz("UCCSD",add_singles=False)
#
#    E = tq.ExpectationValue(H=H, U=U)
#    tqE = tq.ExpectationValue(H=tqH, U=tqU)
#    res = tq.minimize(E, silent=True)
#    tqres = tq.minimize(tqE, silent=True)
#    wfn = tq.simulate(U, variables=res.variables)
#    tqwfn = tq.simulate(tqU, variables=tqres.variables)
#    assert numpy.isclose(res.energy, tqres.energy, atol=1.e-4)
#    assert numpy.isclose(abs(tqwfn.inner(wfn)), 1.0, atol=1.e-4)



import tequila as tq
from tequila import QCircuit,QubitWaveFunction,QubitHamiltonian
from tequila.objective.objective import Variables
from tequila.quantumchemistry.qc_base import QuantumChemistryBase as tqMolecule
import numpy as np
from scipy.linalg import logm, eigh
import itertools
import math
from typing import List,Tuple,Union
from sunrise.orbital_correlation.density_matrix_utils import *
from scipy.optimize import minimize as scp_min

# Quantum entropy S(rho)
def quantum_entropy(rho:np.ndarray)->float:
    """
    Compute the quantum entropy S(rho).

    Parameters:
        rho (ndarray): Density matrix rho (Hermitian, positive semidefinite, trace = 1).

    Returns:
        float: The quantum entropy S(rho).
    """
    # Ensure rho is a numpy array
    rho = np.array(rho, dtype=np.complex128)

    # Validate the input density matrix
    if not np.allclose(rho, rho.conj().T):
        raise ValueError("rho must be Hermitian.")
    if not np.isclose(np.trace(rho), 1):
        raise ValueError("Trace of rho must be 1.")
    if np.any(np.linalg.eigvalsh(rho).round() < 0):
        raise ValueError("rho must be positive semidefinite.")

    # Compute directly
    # log_rho = logm(rho)
    # entropy = -np.trace(rho @ log_rho).real

    # Compute through eigenvalues
    rho_evals, rho_evecs = eigh(rho)
    rho_evals = np.clip(rho_evals, 1e-12, None)
    log_rho = rho_evecs @ np.diag(np.log(rho_evals)) @ rho_evecs.conj().T
    entropy = -np.trace(rho @ log_rho).real

    return entropy

# Quantum relative entropy S(rho||sigma)
def quantum_relative_entropy(rho:np.ndarray, sigma:np.ndarray)->float:
    """
    Compute the quantum relative entropy S(rho || sigma).

    Parameters:
        rho (ndarray): Density matrix rho (Hermitian, positive semidefinite, trace = 1).
        sigma (ndarray): Density matrix sigma (Hermitian, positive semidefinite, trace = 1).

    Returns:
        float: The quantum relative entropy S(rho || sigma).
    """
    # Ensure rho and sigma are numpy arrays
    rho = np.array(rho, dtype=np.complex128)
    sigma = np.array(sigma, dtype=np.complex128)

    # Validate the input density matrices
    if not np.allclose(rho, rho.conj().T):
        raise ValueError("rho must be Hermitian.")
    if not np.allclose(sigma, sigma.conj().T):
        raise ValueError("sigma must be Hermitian.")
    if not np.isclose(np.trace(rho), 1):
        raise ValueError("Trace of rho must be 1.")
    if not np.isclose(np.trace(sigma), 1):
        raise ValueError("Trace of sigma must be 1.")
    if np.any(np.linalg.eigvalsh(rho).round() < 0):
        raise ValueError("rho must be positive semidefinite.")
    if np.any(np.linalg.eigvalsh(sigma).round() < 0):
        raise ValueError("sigma must be positive semidefinite.")

    # Add some noise to make it positive semidefinite
    # epsilon = 1e-10
    # rho = (rho + epsilon * np.eye(sigma.shape[0])).real
    # sigma = (sigma + epsilon * np.eye(sigma.shape[0])).real

    # Compute directly
    # log_rho = logm(rho)
    # log_sigma = logm(sigma)
    # relative_entropy = np.trace(rho @ (log_rho - log_sigma)).real

    # Compute through eigenvalues
    rho_evals, rho_evecs = eigh(rho)
    sigma_evals, sigma_evecs = eigh(sigma)
    rho_evals = np.clip(rho_evals, 1e-12, None)
    sigma_evals = np.clip(sigma_evals, 1e-12, None)
    log_rho = rho_evecs @ np.diag(np.log(rho_evals)) @ rho_evecs.conj().T
    log_sigma = sigma_evecs @ np.diag(np.log(sigma_evals)) @ sigma_evecs.conj().T
    relative_entropy = np.trace(rho @ (log_rho - log_sigma)).real

    return relative_entropy

def mutual_info_simple(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, orb_b:int=1, PSSR:bool=False, NSSR:bool=False)->float:
    rho_a = compute_one_orb_rdm(mol, circuit, variables, initial_state, orb_a)
    S_a = quantum_entropy(rho_a)
    rho_b = compute_one_orb_rdm(mol, circuit, variables, initial_state, orb_b)
    S_b = quantum_entropy(rho_b)
    rho_ab = compute_two_orb_rdm(mol, circuit, variables, initial_state, p_orb=orb_a, q_orb=orb_b, PSSR=PSSR, NSSR=NSSR)
    S_ab = quantum_entropy(rho_ab)

    return S_a + S_b - S_ab # there might be a 0.5 depending to convention

def one_orb_mutual_info(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, PSSR:bool=False, NSSR:bool=False):
    rho_a = compute_one_orb_rdm(mol, circuit, variables, initial_state, orb_a)
    if PSSR:
        rho_a_evals, rho_a_evecs = eigh(rho_a)
        I = (rho_a_evals[0]+rho_a_evals[3])*np.log(rho_a_evals[0]+rho_a_evals[3]) + \
            (rho_a_evals[1]+rho_a_evals[2])*np.log(rho_a_evals[1]+rho_a_evals[2]) - \
            2*(rho_a_evals[0]*np.log(rho_a_evals[0])+rho_a_evals[1]*np.log(rho_a_evals[1])+\
               rho_a_evals[2]*np.log(rho_a_evals[2])+rho_a_evals[3]*np.log(rho_a_evals[3]))
    elif NSSR:
        rho_a_evals, rho_a_evecs = eigh(rho_a)
        I = rho_a_evals[0]*np.log(rho_a_evals[0]) + \
            (rho_a_evals[1]+rho_a_evals[2])*np.log(rho_a_evals[1]+rho_a_evals[2]) + \
            rho_a_evals[3]*np.log(rho_a_evals[3]) - \
            2*(rho_a_evals[0]*np.log(rho_a_evals[0])+rho_a_evals[1]*np.log(rho_a_evals[1])+\
               rho_a_evals[2]*np.log(rho_a_evals[2])+rho_a_evals[3]*np.log(rho_a_evals[3]))
    else:
        I = 2*one_orb_entanglement(mol, circuit, variables, initial_state, orb_a=orb_a)

    return I

def total_mutual_info(mol, circuit=None, variables=None, initial_state=0, orbs=None, PSSR=False, NSSR=False):
    """
    Compute the mutual information between specified orbitals and the rest of the system.
    """

    if not orbs:
        orbs = list(range(mol.n_orbitals))

    one_entropies = []
    for orb in orbs:
        one_rdm = compute_one_orb_rdm(mol, circuit, variables, initial_state, orb)
        one_entropies.append(quantum_entropy(one_rdm))
    one_entropy = sum(one_entropies)

    state = tq.simulate(circuit, variables=variables, initial_state=initial_state)
    rho = tq.paulis.Projector(state).to_matrix().real
    system_entropy = quantum_entropy(rho)

    return one_entropy - system_entropy

def one_orb_entanglement(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, PSSR:bool=False, NSSR:bool=False)->float:
    rho_a = compute_one_orb_rdm(mol, circuit, variables, initial_state, orb_a)
    if PSSR==True:
        rho_a_evals, rho_a_evecs = eigh(rho_a)
        # Eq.(29) https://doi.org/10.1021/acs.jctc.0c00559
        plnp = [(rho_a_evals[i]+rho_a_evals[-1-i])*np.log(rho_a_evals[i]+rho_a_evals[-1-i]) if not np.isclose(rho_a_evals[i]+rho_a_evals[-1-i], 0, atol=1.e-6) else 0.0 for i in range(len(rho_a_evals)//2)]
        xlnx =[i * np.log(i) if not np.isclose(i, 0, atol=1.e-6) else 0.0 for i in rho_a_evals]
        E = (plnp[0]+plnp[1] ) - (xlnx[0]+xlnx[1]+xlnx[2]+xlnx[3])
    elif NSSR==True:
        rho_a_evals, rho_a_evecs = eigh(rho_a)
        # Eq.(29) https://doi.org/10.1021/acs.jctc.0c00559
        xlnx =[i * np.log(i) if not np.isclose(i, 0, atol=1.e-6) else 0.0 for i in rho_a_evals]
        E = (rho_a_evals[1]+rho_a_evals[2])*np.log(rho_a_evals[1]+rho_a_evals[2]) if not np.isclose(rho_a_evals[1]+rho_a_evals[2], 0, atol=1.e-6) else 0.0
        E -= (xlnx[1] + xlnx[2])
    else:
        S_a = quantum_entropy(rho_a)
        # assert np.isclose(S_a,S_b)
        E = S_a

    return E



def func(x, d, rho):
    """
    Computes S(rho || sigma) where sigma is constructed from x.
    Probabilities are derived from the norms of the vectors in x.
    """
    x_reshaped = x.reshape(2*d, 4)
    components = []
    
    # Extract vectors
    vecs_A = x_reshaped[0::2]
    vecs_B = x_reshaped[1::2]
    
    # Calculate norms
    nA = np.linalg.norm(vecs_A, axis=1)
    nB = np.linalg.norm(vecs_B, axis=1)
    
    # Calculate weights: p_k = |v_A|^2 * |v_B|^2
    # Add epsilon to avoid division by zero
    raw_weights = (nA**2) * (nB**2)
    total_weight = np.sum(raw_weights)
    
    if total_weight < 1e-12:
        # Fallback for zero vectors
        probs = np.ones(d) / d
    else:
        probs = raw_weights / total_weight

    for k in range(d):
        # Normalize vectors for the state creation
        # (create_separable_mixed_state handles normalization, but we need norms for probs)
        vA = vecs_A[k]
        vB = vecs_B[k]
        components.append((probs[k], vA, vB))

    # Construct Sigma
    sigma = create_separable_mixed_state(components)
    sigma = change_basis(sigma, 'to_molecular')
    
    return quantum_relative_entropy(rho=rho, sigma=sigma)

def relative_entropy_gradient_matrix(rho, sigma, epsilon=1e-12):
    """
    Computes the matrix gradient of -Tr(rho log sigma) w.r.t sigma
    using the Daleckii-Krein formula (divided differences).
    """
    # 1. Diagonalize sigma
    s_evals, s_evecs = eigh(sigma)
    
    # Clip eigenvalues to avoid log(0) errors
    s_evals = np.clip(s_evals, epsilon, None)
    
    # 2. Rotate rho to sigma basis: rho_rot = U^dag @ rho @ U
    rho_rot = s_evecs.conj().T @ rho @ s_evecs
    
    # 3. Prepare differences
    lambda_j = s_evals[:, None] 
    lambda_k = s_evals[None, :]
    
    lin_diff = lambda_j - lambda_k
    mask = np.abs(lin_diff) < epsilon
    
    D = np.zeros_like(rho)
    
    # --- CASE 1: Non-Degenerate ---
    log_lambda = np.log(s_evals)
    log_diff = log_lambda[:, None] - log_lambda[None, :]
    D[~mask] = log_diff[~mask] / lin_diff[~mask]
    
    # --- CASE 2: Degenerate (Limit Case) ---
    rows, cols = np.nonzero(mask)
    D[rows, cols] = 1.0 / s_evals[rows]
    
    # 4. Element-wise multiply with rotated rho
    grad_rot = -rho_rot * D
    
    # 5. Rotate back to original basis
    grad_sigma = s_evecs @ grad_rot @ s_evecs.conj().T
    
    return grad_sigma.real

def gradient_func(x, d, rho):
    x_reshaped = x.reshape(2*d, -1)
    
    # --- 1. Forward Pass ---
    vecs_A = x_reshaped[0::2]
    vecs_B = x_reshaped[1::2]
    
    nA = np.linalg.norm(vecs_A, axis=1)
    nB = np.linalg.norm(vecs_B, axis=1)
    
    psi_A = vecs_A / (nA[:, None] + 1e-12)
    psi_B = vecs_B / (nB[:, None] + 1e-12)
    
    raw_weights = nA**2 * nB**2
    total_weight = np.sum(raw_weights)
    probs = raw_weights / total_weight
    
    components = []
    for k in range(d):
        components.append((probs[k], psi_A[k], psi_B[k]))
        
    # Create sigma in computational basis
    sigma_sep = create_separable_mixed_state(components)
    
    # Transform to molecular basis to compute gradient against rho
    sigma_mol = change_basis(sigma_sep, 'to_molecular')
    
    # --- 2. Matrix Gradient in Molecular Basis ---
    grad_sigma_mol = relative_entropy_gradient_matrix(rho, sigma_mol)
    
    # --- 3. Transform Gradient back to Computational Basis ---
    # This replaces the inverse permutation logic
    grad_sigma_comp = change_basis(grad_sigma_mol, 'to_computational')
        
    # --- 4. Chain Rule for Norm-Weighted Params ---
    grad_x = np.zeros_like(x_reshaped)
    dimA, dimB = psi_A.shape[1], psi_B.shape[1]
    
    # Use the computational basis gradient for backprop
    G_tensor = grad_sigma_comp.reshape(dimA, dimB, dimA, dimB)
    
    traces = np.zeros(d)
    grad_projectors_A = []
    grad_projectors_B = []
    
    for k in range(d):
        # Contract for Projector Gradients
        g_PA = probs[k] * np.einsum('ikjl,k,l->ij', G_tensor, psi_B[k], psi_B[k].conj()).real
        g_PB = probs[k] * np.einsum('ikjl,i,j->kl', G_tensor, psi_A[k], psi_A[k].conj()).real
        
        grad_projectors_A.append(g_PA)
        grad_projectors_B.append(g_PB)
        
        # Contract for Weight Gradients
        term = np.einsum('ij,ij->', g_PA, np.outer(psi_A[k], psi_A[k].conj())).real
        traces[k] = term / probs[k] if probs[k] > 1e-10 else 0

    avg_trace = np.dot(probs, traces)
    dL_dw = (traces - avg_trace) / total_weight
    
    for k in range(d):
        # Projector gradient part
        def unnorm_grad(g_psi, psi, n):
            return (g_psi - np.dot(psi.conj(), g_psi).real * psi) / (n + 1e-12)

        g_psiA = 2 * grad_projectors_A[k] @ psi_A[k]
        g_psiB = 2 * grad_projectors_B[k] @ psi_B[k]
        
        term1_A = unnorm_grad(g_psiA, psi_A[k], nA[k])
        term1_B = unnorm_grad(g_psiB, psi_B[k], nB[k])
        
        # Weight gradient part
        d_weight_dA = dL_dw[k] * 2 * nA[k] * (nB[k]**2)
        d_weight_dB = dL_dw[k] * 2 * nB[k] * (nA[k]**2)
        
        term2_A = d_weight_dA * psi_A[k]
        term2_B = d_weight_dB * psi_B[k]
        
        grad_x[2*k] = term1_A + term2_A
        grad_x[2*k+1] = term1_B + term2_B

    return grad_x.flatten()

def get_classical_diagonal_guess(rho):
    """
    Generates the Classical Diagonal state parameters from a 16x16 Density Matrix.
    """
    # 1. Define the 4 basis vectors for a single orbital
    basis_vectors = np.eye(4) 
    
    states = []
    x0 = []

    # FIX: The input rho is in the Molecular Basis.
    # We must rotate it to Computational Basis to interpret index k as (idx_A * 4 + idx_B).
    rho_computational = change_basis(rho, direction='to_computational')

    # 2. Iterate through the diagonal of the COMPUTATIONAL basis matrix
    for k in range(16):
        prob = np.real(rho_computational[k, k])
        
        if prob > 1e-6:
            # Now k correctly corresponds to the computational basis index
            idx_A = k // 4 
            idx_B = k % 4 
            
            vec_A = basis_vectors[idx_A] 
            vec_B = basis_vectors[idx_B] 
            
            states.append((prob, vec_A, vec_B))
            
            weighted_vec_A = vec_A * np.sqrt(prob)
            weighted_vec_B = vec_B * np.sqrt(prob)
            
            x0.extend(weighted_vec_A)
            x0.extend(weighted_vec_B)

    return states, np.array(x0)

def two_orbs_entanglement(
    mol: tq.Molecule, 
    circuit: tq.QCircuit = None, 
    variables = None, 
    initial_state = None, 
    orb_a: int = 0, 
    orb_b: int = 1, 
    PSSR: bool = False, 
    NSSR: bool = False,
    min_components: int = 8,
    silent: bool = True
) -> float:
    """
    Computes the Relative Entropy of Entanglement E_R(rho) for a pair of molecular orbitals.
    
    Automates the process of:
    1. Computing the 2-orbital RDM (rho_ab).
    2. Generating a classical guess for the separable state.
    3. Padding the guess to ensure enough degrees of freedom (d >= min_components).
    4. Minimizing S(rho || sigma) using analytic gradients.
    
    Args:
        mol: Tequila Molecule object.
        circuit: The ansatz circuit (optional).
        variables: Optimized variables for the circuit (optional).
        initial_state: Initial state for RDM computation (optional).
        orb_a, orb_b: Indices of the two orbitals to test.
        PSSR, NSSR: Flags passed to compute_two_orb_rdm.
        min_components: Minimum number of separable components (d) to enforce. 
                        Prevents getting stuck in pure-state local minima.
        silent: If False, prints optimization progress.

    Returns:
        float: The minimized relative entropy (entanglement).
    """
    
    # 1. Compute the Reduced Density Matrix (rho_ab)
    # Note: Ensure your compute_two_orb_rdm handles the 'variables' and 'circuit' mapping correctly
    # If variables are provided, usually we map them to the circuit before passing, 
    # but here we pass them through as requested.
    if not silent: 
        print(f"Computing RDM for orbitals ({orb_a}, {orb_b})...")
        
    # Map variables if provided and circuit exists, otherwise pass as is
    # (Adjust this line based on how compute_two_orb_rdm expects the circuit)
    if circuit is not None and variables is not None:
        circuit = circuit.map_variables(variables)

    rho = compute_two_orb_rdm(mol, circuit, variables, initial_state, p_orb=orb_a, q_orb=orb_b, PSSR=PSSR, NSSR=NSSR)

    # 2. Generate Initial Guess (Classical Diagonal)
    states0, x0 = get_classical_diagonal_guess(rho)
    current_d = len(states0)
    
    # 3. Robustness Padding
    # If the guess is too simple (e.g., just the HF state), we need to add "empty" components
    # to allow the optimizer to find a mixed state solution.
    d = current_d
    if current_d < min_components:
        if not silent:
            print(f"  Padding components from {current_d} to {min_components} for robustness.")
        
        missing_d = min_components - current_d
        # Each component has 2 vectors of size 4 -> 8 parameters per component
        # Initialize with small random noise to break symmetry
        padding = np.random.randn(missing_d * 8) * 0.1 
        x0 = np.concatenate([x0, padding])
        d = min_components

    # 4. Symmetry Breaking
    # Add slight noise to the whole vector to avoid saddle points
    x0 = x0 + np.random.randn(*x0.shape) * 0.05

    # 5. Run Optimization
    args = (d, rho)
    
    if not silent:
        print("  Minimizing relative entropy...")

    result = scp_min(
        fun=func, 
        x0=x0, 
        args=args, 
        method="L-BFGS-B",  # Efficient gradient-based method
        jac=gradient_func,  # Analytic Gradient
        options={
            "maxiter": 2000, 
            "ftol": 1e-9, 
            "disp": False
        }
    )

    if not result.success and not silent:
        print(f"  Warning: Optimization finished with status: {result.message}")

    # Return the entanglement value
    return result.fun

def two_orbs_quantum_correlation(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, orb_b:int=1, PSSR:bool=False, NSSR:bool=False)->float:
    
    rho_ab = compute_two_orb_rdm(mol, circuit, variables, initial_state, p_orb=orb_a, q_orb=orb_b, PSSR=PSSR, NSSR=NSSR)
    chi = np.diag(np.diag(rho_ab))
        
    return quantum_relative_entropy(rho_ab, chi)

def two_orbs_classical_correlation(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, orb_b:int=1, PSSR:bool=False, NSSR:bool=False)->float:
    
    rho_ab = compute_two_orb_rdm(mol, circuit, variables, initial_state, p_orb=orb_a, q_orb=orb_b, PSSR=PSSR, NSSR=NSSR)
    chi = np.diag(np.diag(rho_ab))

    o1 = compute_one_orb_rdm(mol, circuit, one_orb=orb_a)
    o2 = compute_one_orb_rdm(mol, circuit, one_orb=orb_b)
    pi = np.kron(o1,o2)
    pi = change_basis(pi, 'to_molecular')
        
    return quantum_relative_entropy(chi, pi)

def two_orbs_mutual_info(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, orb_a:int=0, orb_b:int=1, PSSR:bool=False, NSSR:bool=False)->float:
    
    rho_ab = compute_two_orb_rdm(mol, circuit, variables, initial_state, p_orb=orb_a, q_orb=orb_b, PSSR=PSSR, NSSR=NSSR)
    o1 = compute_one_orb_rdm(mol, circuit, one_orb=orb_a)
    o2 = compute_one_orb_rdm(mol, circuit, one_orb=orb_b)
    pi = np.kron(o1,o2)
    pi = change_basis(pi, 'to_molecular')
        
    return quantum_relative_entropy(rho_ab, pi)



import tencirchem as tcc
from tencirchem.static.ci_utils import get_ci_strings
from sunrise.expval.tcc_engine.braket import EXPVAL
from ..fermionic_operations.circuit import FCircuit
from tequila import TequilaException,Molecule,QubitWaveFunction,simulate,Variable,Objective,assign_variable,QubitHamiltonian
from tequila import grad as tq_grad
from tequila.objective.objective import Variables,FixedVariable
from tequila.quantumchemistry.chemistry_tools import NBodyTensor
from tequila.quantumchemistry import qc_base
from tequila.utils.bitstrings import BitString, BitNumbering
from numbers import Number
from numpy import ceil,argwhere,pi,prod,eye,zeros,isclose,allclose
from pyscf.gto import Mole
from pyscf.scf import RHF
from sunrise.expval.pyscf_molecule import from_tequila
from copy import deepcopy
from typing import Union,List,Tuple,Callable
from collections import defaultdict
from openfermion import FermionOperator
from openfermion.transforms import jordan_wigner
from openfermion.transforms.opconversions.term_reordering import reorder
from openfermion.utils.indexing import up_then_down

class TCCBraket:
    def __init__(self,bra:Union[FCircuit,None]=None,ket:Union[FCircuit,None]=None,operator:Union[str,FermionOperator,List[FermionOperator]]=None,backend_kwargs:dict={},*args,**kwargs):
        self.operator = None
        if 'engine' in backend_kwargs:
            engine = backend_kwargs['engine']
            backend_kwargs.pop('engine')
        else: engine = 'pyscf'
        if 'backend' in  backend_kwargs:
            tcc.set_backend(backend_kwargs['backend'])
            backend_kwargs.pop('backend')
        else: tcc.set_backend('numpy')
        if 'dtype' in backend_kwargs:
            tcc.set_dtype(backend_kwargs['dtype'])
            backend_kwargs.pop('dtype')
        
        if 'circuit' in kwargs:
            circuit = kwargs['circuit']
            kwargs.pop('circuit')
            if ket is not None:
                raise TequilaException('Two circuits provided?')
            else:
                ket = circuit
        if 'U' in kwargs:
            U = kwargs['U']
            kwargs.pop('U')
            if ket is not None:
                raise TequilaException('Two circuits provided?')
            else:
                ket = U
        if 'H' in kwargs:
            H = kwargs['H']
            kwargs.pop('H')
            if operator is not None:
                raise TequilaException('Two operators provided?')
            else:
                operator = H
        if 'mol' in kwargs:
            if 'molecule' in kwargs and kwargs['molecule']:
                raise TequilaException("Two molecules provided?")
            kwargs['molecule'] = kwargs['mol']
            kwargs.pop('mol')

        run_hf = (bra is None or bra.initial_state is None) and (ket is None or ket.initial_state is None)   
        if 'molecule' in kwargs and kwargs['molecule']:
            molecule = kwargs['molecule']
            kwargs.pop('molecule')
            if isinstance(molecule,qc_base.QuantumChemistryBase):
                mo_coeff = molecule.integral_manager.orbital_coefficients 
                aslst = [i.idx_total for i in molecule.integral_manager.active_orbitals]
                active_space = (molecule.n_electrons,molecule.n_orbitals)
                if allclose(mo_coeff,eye(len(mo_coeff))): #idea when initialized the molecule from integrals, the mo_coeff are setted to identity and the provided integral are saved on the place of the Atomic Integrals
                    e_core,int1e,int2e = molecule.get_integrals()
                    int2e = int2e.reorder('c').elems
                    self.BK:EXPVAL = EXPVAL.from_integral(int1e=int1e, int2e=int2e,n_elec=molecule.n_electrons, e_core=e_core,mo_coeff=mo_coeff,init_method="zeros",engine=engine,run_hf= run_hf, run_mp2= False, run_ccsd= False, run_fci= False,**backend_kwargs)
                else:
                    molecule = from_tequila(molecule)
                    self.BK:EXPVAL = EXPVAL(mol=molecule,run_hf= run_hf, run_mp2= False, run_ccsd= False, run_fci= False,init_method="zeros",aslst=aslst,active_space=active_space,engine=engine,mo_coeff=mo_coeff,**backend_kwargs)
            elif isinstance(molecule,Mole):
                mf = RHF(mol=molecule)
                mo_coeff =mf.mo_coeff
                aslst = [*range(molecule.nao_nr_range)]
                active_space = (molecule.nelectron,molecule.nao_nr)
                self.BK:EXPVAL = EXPVAL(mol=molecule,run_hf= run_hf, run_mp2= False, run_ccsd= False, run_fci= False,init_method="zeros",aslst=aslst,active_space=active_space,engine=engine,mo_coeff=mo_coeff,**backend_kwargs)
        elif 'integral_manager' in kwargs and 'parameters' in kwargs:
            integral = kwargs['integral_manager']
            params = kwargs['parameters']
            kwargs.pop('integral_manager')
            kwargs.pop('parameters')
            mo_coeff = integral.orbital_coefficients
            molecule = Molecule(parameters=params,integral_manager=integral)
            aslst = [i.idx_total for i in integral.active_orbitals]
            active_space = (molecule.n_electrons,molecule.n_orbitals)
            molecule = from_tequila(molecule)
            self.BK:EXPVAL = EXPVAL(mol=molecule,run_hf= run_hf, aslst=aslst,active_space=active_space,run_mp2= False, run_ccsd= False, run_fci= False,init_method="zeros",engine=engine,mo_coeff=mo_coeff,**backend_kwargs)
        else:
            int1e = None
            int2e = None
            e_core = None
            mo_coeff = None
            ovlp = None
            if "int1e"  in kwargs:
                int1e = kwargs['int1e']
                kwargs.pop('int1e')
            elif "one_body_integrals"  in kwargs:
                int1e = kwargs['one_body_integrals']
                kwargs.pop('one_body_integrals')
            elif "h"  in kwargs:
                int1e = kwargs['h']
                kwargs.pop('h')
            if 'int2e' in kwargs:
                int2e = kwargs['int2e']
                kwargs.pop('int2e')
            elif 'two_body_integrals' in kwargs:
                int2e = kwargs['two_body_integrals']
                kwargs.pop('two_body_integrals')
            elif 'g' in kwargs:
                int2e = kwargs['g']
                kwargs.pop('g')
            if isinstance(int2e,NBodyTensor):
                int2e = int2e.elems
            if 'e_core' in kwargs:
                e_core = kwargs['e_core']
                kwargs.pop('e_core')
            elif 'constant_term' in kwargs:
                e_core = kwargs['constant_term']
                kwargs.pop('constant_term')
            elif 'constant' in kwargs:
                e_core = kwargs['constant']
                kwargs.pop('constant')
            elif 'c' in kwargs:
                e_core = kwargs['c']
                kwargs.pop('c')    
            else: e_core = 0.
            # if 'mo_coeff' in kwargs:
            #     mo_coeff = kwargs['mo_coeff']
            #     kwargs.pop('mo_coeff')
            # elif 'orbital_coefficients' in kwargs:
            #     mo_coeff = kwargs['orbital_coefficients']
            #     kwargs.pop('orbital_coefficients')
            mo_coeff = eye(len(int1e)) #IDEA The intengrals refer to the MO, so the active space stuff is kept out of the braket 
            if 'ovlp' in kwargs:
                ovlp = kwargs['ovlp']
                kwargs.pop('ovlp')
            elif 'overlap_integrals' in kwargs:
                ovlp = kwargs['overlap_integrals']
                kwargs.pop('overlap_integrals')
            elif 's' in kwargs:
                ovlp = kwargs['s']
                kwargs.pop('s')
            if 'n_elec' in kwargs:
                n_elec=kwargs['n_elec']
                kwargs.pop('n_elec')
            elif 'n_electrons' in kwargs: 
                n_elec=kwargs['n_elec']
                kwargs.pop('n_elec')
            elif ket is not None and ket.init_state is not None:
                if isinstance(ket.initial_state._state,dict):
                    n_elec = bin([*ket.initial_state._state.keys()][0])[2:].count('1')
                else:
                    n_elec = bin(argwhere(ket.init_state._state>1.e-6)[0][0])[2:].count('1')
            else:
                raise TequilaException("No manner of defining the amount of electrons provided")
            if all([i is not None for i in[int2e,int1e,mo_coeff]]):
                if isinstance(int2e,NBodyTensor):
                    int2e = int2e.reorder('chem').elems
                self.BK:EXPVAL = EXPVAL.from_integral(int1e=int1e, int2e=int2e,n_elec= n_elec, e_core=e_core,ovlp=ovlp,mo_coeff=mo_coeff,init_method="zeros",engine=engine,run_hf= run_hf, run_mp2= False, run_ccsd= False, run_fci= False,**backend_kwargs)
            else:
                raise TequilaException('Not enough molecular data provided')
        if ket is not None:
            self.ket = ket
        if bra is not None:
            self.bra = bra 
        self.opt_res = defaultdict(None)
        if 'name' in kwargs:
            self._name = kwargs['name']
        else: self._name = 'Expectation Value' if self.is_diagonal else "Transition Value"
        if isinstance(operator,str) and operator == 'I':
            self._name = 'Transition Element'
        if operator is not None:
            self.operator = self.build_operator(operator)
        
    def minimize(self,**kwargs)->float:
        if 'init_guess_bra' in kwargs:
            self.init_guess_bra = kwargs['init_guess_bra']
        if "init_guess_ket" in kwargs:
            self.init_guess_ket = kwargs['init_guess_ket']
        if "init_guess" in kwargs:
            self.init_state = kwargs["init_guess"]
        e = self.BK.kernel()
        if self.BK.opt_res is not None:
            self.opt_res = deepcopy(self.BK.opt_res)
            self.opt_res.x = [-2*i for i in self.opt_res.x] #translating to tq
            return self.BK.opt_res.e
        else: 
            self.opt_res['e'] = e
            return e

    def __call__(self, variables:Union[list,dict]={}, *args, **kwargs) -> float:
        return self.simulate(variables=variables)

    def simulate(self,variables:Union[list,dict]=None)->float:
        if isinstance(variables,Variables):
            variables = variables.store
        if isinstance(variables,dict):
            v: dict = deepcopy(self.variables)
            if v is None:
                v = variables
            else:
                v.update(variables)
            tvars: list = deepcopy(self.BK.total_variables)
            variables:list = [map_variables(x,v) for x in tvars]
        if variables is None:
            return self.BK.expval(hamiltonian=self.operator)
        return self.BK.expval(angles=[-0.5*i for i in variables],hamiltonian=self.operator)

    def extract_variables(self) -> list:
        """
        Extract all variables on which the objective depends
        :return: List of all Variables
        """
        v = []
        for d in self.params:
            if isinstance(d,FixedVariable):
                continue
            elif hasattr(d,'extract_variables'):
                v.extend(d.extract_variables())
            else:
                v.extend(d)
        unique = []
        for i in v:
            if i not in unique:
                unique.append(i)
        return unique

    def grad(self,variable:Variable = None)->Objective:
        def apply_phase(braket: TCCBraket,exct:List[Tuple[int]],variable,ket:bool=True,p0sign:bool=True)->Objective:
            '''
            braket: TCC object to modify
            exct: Excitation indices on the tequila format [(0,2),(1,3),...] to which 
                  apply the phase shift.
            ket: If true it will be applied the phase on the ket side, bra otherwise
            posing: it True: +pi, False: -Pi, correspond to the U0(\pm) not the actual sign implementation
            '''
            p0 = []
            s = {True:+1,False:-1} 

            for idx in exct:
                p0.append((idx[0],idx[0]))
                p0.append((idx[1],idx[1]))
            braket._name = 'Gradient'
            if ket:
                if braket.is_diagonal: 
                    k = deepcopy(braket.ket)
                    v = deepcopy(braket.params_ket)
                    braket.bra = deepcopy(k)
                    braket.variables_bra = deepcopy(v)
                    idx = k.index(exct)
                    ph = tq_grad(v[idx],variable) if not isinstance(v[idx],Union[FixedVariable,Number,Variable]) else 1
                    v[idx] +=  s[ket]*pi/2 
                    for p in reversed(p0):
                        k.insert(idx,[p])
                        v.insert(idx,assign_variable(s[not p0sign]*pi))
                    braket.ket = k
                    braket.variables_ket = v
                else:
                    k = deepcopy(braket.ket)
                    v = deepcopy(braket.params_ket)
                    if exct not in k:
                        return 0.
                    idx = k.index(exct)
                    ph = tq_grad(1*v[idx],variable) if not isinstance(v[idx],Union[FixedVariable,Number,Variable]) else 1
                    v[idx] +=  s[ket]*pi/2
                    for p in reversed(p0):
                        k.insert(idx,[p])
                        v.insert(idx,assign_variable(s[not p0sign]*pi)) 
                    braket.ket = k
                    braket.variables_ket = v
            else: 
                if braket.is_diagonal: 
                    k = deepcopy(braket.ket)
                    v = deepcopy(braket.params_ket)
                    idx = k.index(exct)
                    ph = tq_grad(v[idx],variable) if not isinstance(v[idx],Union[FixedVariable,Number,Variable]) else 1
                    v[idx] +=  s[ket]*pi/2 
                    for p in reversed(p0):
                        k.insert(idx,[p])
                        v.insert(idx,assign_variable(s[not p0sign]*pi)) 
                    braket.bra = k
                    braket.variables_bra = v
                else:
                    k = deepcopy(braket.bra)
                    v = deepcopy(braket.params_bra)
                    if exct not in k:
                        return 0.
                    idx = k.index(exct)
                    ph = tq_grad(v[idx],variable) if not isinstance(v[idx],Union[FixedVariable,Number,Variable]) else 1
                    v[idx] +=  s[ket]*pi/2
                    for p in reversed(p0):
                        k.insert(idx,[p])
                        v.insert(idx,assign_variable(s[not p0sign]*pi)) 
                    braket.bra = k
                    braket.variables_bra = v
            return s[ket]*s[(len(p0)//2)%2]*s[p0sign]*ph*Objective([braket]) #TODO: Check this correct
        if variable is None:
            # None means that all components are created
            variables = self.extract_variables()
            result = {}

            if len(variables) == 0:
                raise TequilaException("Error in gradient: Objective has no variables")

            for k in variables:
                assert k is not None
                result[k] = self.grad(k)
            return result
        else:
            variable = assign_variable(variable)
        if variable not in self.extract_variables():
            return 0.
        if 'civector' in self.BK.engine:
            try:
                _,grad = self.BK.expval_and_grad(angles=variable)
                return grad
            except: raise TequilaException("For civector engine it doesn't work out P0 approach for gradient, try better the TCCBraket.minimize() or change to other engine")
        ex_ops = self.param_to_ex_ops[variable]
        g = 0
        for exct in ex_ops:
            g +=apply_phase(braket=deepcopy(self),exct=exct,variable=variable,ket=True,p0sign=True) 
            # g +=apply_phase(braket=deepcopy(self),exct=exct,variable=variable,ket=True,p0sign=False) #wfn always real for tcc 
            g +=apply_phase(braket=deepcopy(self),exct=exct,variable=variable,ket=False,p0sign=True)
            # g +=apply_phase(braket=deepcopy(self),exct=exct,variable=variable,ket=False,p0sign=False) #wfn always real for tcc
        return 0.5*g

    @property
    def energy(self)->float:
        if isinstance(self.opt_res,dict):
            return self.opt_res['e']
        else:
            return self.opt_res.e

    @property
    def bra(self):
        """
        Excitation operators applied to the bra.
        """
        return indices_tcc_to_tq(self.BK.ex_ops_bra)  
    
    @bra.setter
    def bra(self, bra:Union[List[List[Tuple[int]]],FCircuit]):
        '''
        Expected FCircuit or indices in [[(0,2),(1,3),...],[(a,b),(c,d),...],...] Format (Upthendown order)
        '''
        if isinstance(bra,FCircuit):
            if bra.initial_state is not None:
                ini_bra = init_state_from_wavefunction(bra.initial_state)
                for i in ini_bra:
                    i[0] = self.BK.get_addr(i[0])
                init = [0] * self.BK.civector_size
                for i in ini_bra:
                    init[i[0]] = i[1]
                self.init_state_bra = init
            self.variables_bra = bra.variables
            bra = bra.to_upthendown(len(self.BK.aslst))
            self.BK.ex_ops_bra,_,_ =indices_tq_to_tcc(bra.extract_indices())
        else:
            bra,params_bra,_ = indices_tq_to_tcc(bra)
            if self.variables_bra is None:
                self.variables_bra = params_bra
            self.BK.ex_ops_bra = bra
    
    @property
    def ket(self):
        """
        Excitation operators applied to the ket.
        """
        return indices_tcc_to_tq(self.BK.ex_ops_ket)
    
    @ket.setter
    def ket(self, ket:Union[List[List[Tuple[int]]],FCircuit]):
        '''
        Expected FCircuit or indices in [[(0,2),(1,3),...],[(a,b),(c,d),...],...] Format (Upthendown order)
        '''
        if isinstance(ket,FCircuit):
            if ket.initial_state is not None:
                ini_ket = init_state_from_wavefunction(ket.initial_state)
                for i in ini_ket:
                    i[0] = self.BK.get_addr(i[0])
                init = [0] * self.BK.civector_size
                for i in ini_ket:
                    init[i[0]] = i[1]
                self.init_state_ket = init
            self.variables_ket = ket.variables
            ket = ket.to_upthendown(len(self.BK.aslst))
            self.BK.ex_ops_ket,_,_ =indices_tq_to_tcc(ket.extract_indices())
        else:
            ket,params_ket,_ = indices_tq_to_tcc(ket)
            if self.variables_ket is None:
                self.variables_ket = params_ket
            self.BK.ex_ops_ket = ket

    @property
    def variables_bra(self) -> dict:
        """Tequila Circuit Bra variables."""
        bar:dict = deepcopy(self.BK.var_to_param_bra)
        if bar is not None:
            for i in bar.keys(): #Here to tequila
                if isinstance(bar[i],(Number,FixedVariable)):
                    bar[i] = assign_variable(-2*(bar[i]))
        return bar 
    
    @property
    def params_bra(self):
        """TCC Circuit Bra parameters (values after minimization or variables name)."""
        if self.variables_bra is None: return []
        d = {v: k for k, v in self.variables_bra.items()}
        return [assign_variable(-2*v) if isinstance(v,(Number,FixedVariable)) else d[assign_variable(v)] for v in self.BK.params_bra]
    
    @variables_bra.setter
    def variables_bra(self, variables_bra):
        '''
        See TCC variables
        '''
        for idx,i in enumerate(variables_bra):
            if isinstance(assign_variable(i),(Number, FixedVariable)):
                variables_bra[idx]=assign_variable(-0.5*(i%(2*pi)))
        self.BK.params_bra = variables_bra

    @property
    def variables_ket(self):
        """Tequila circuit Ket parameters."""
        bar = deepcopy(self.BK.var_to_param_ket)
        if bar is not None:
            for i in bar.keys(): #to tequila
                if isinstance(bar[i],(Number,FixedVariable)):
                    bar[i] = assign_variable(-2*(bar[i]))
        return bar
    
    @property
    def params_ket(self):
        """TCC Circuit Ket parameters (values after minimization or variables name)."""
        d = {v: k for k, v in self.variables_ket.items()}
        return [assign_variable(-2*v) if isinstance(v,(Number,FixedVariable)) else d[assign_variable(v)] for v in self.BK.params_ket]
    
    @variables_ket.setter
    def variables_ket(self, variables_ket):
        '''
        See TCC variables
        '''
        for idx,i in enumerate(variables_ket):
            if isinstance(assign_variable(i),(Number, FixedVariable)):
                variables_ket[idx]=assign_variable(-0.5*(i%(2*pi)))
        self.BK.params_ket = variables_ket
    
    @property
    def variables(self) -> dict:
        """Tequila circuit variables."""
        bar:dict = deepcopy(self.BK.var_to_param)
        if bar is not None:
            for i in bar.keys(): #Here to tequila 
                if isinstance(bar[i],(Number, FixedVariable)):
                    bar[i] = assign_variable(-2*bar[i])
        return bar 
    
    @property
    def params(self):
        """TCC parameters (values after minimization or variables name)."""
        k = self.params_ket
        if self.BK.params_bra is None:
            return k
        return self.params_bra+k
    
    @variables.setter
    def variables(self, variables):
        """Tequila circuit variables."""
        for idx,i in enumerate(variables):
            if isinstance(assign_variable(i),(Number, FixedVariable)):
                variables[idx]=assign_variable(-0.5*(i%(2*pi)))
        self.BK.params = variables

    @property
    def init_state_bra(self):
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        BK.get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self.BK.init_state_bra
    
    @property
    def init_state_ket(self):
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        BK.get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self.BK.init_state_ket
    
    @property
    def init_state(self):
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        BK.get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self.BK.init_state_bra ,self.BK.init_state_ket

    @init_state_bra.setter
    def init_state_bra(self, init_state_bra):
        self.BK.init_state_bra = init_state_bra
    
    @init_state_ket.setter
    def init_state_ket(self, init_state_ket):
        self.BK.init_state_ket  = init_state_ket

    @init_state.setter
    def init_state(self, init_state):
        self.BK.init_state_bra  = init_state
        self.BK.init_state_ket  = init_state

    @property
    def init_guess(self):
        """
        Initial Angle Value for minimization, default all 0.
        """
        return [-2*i for i in self.BK.init_guess]
    
    @init_guess.setter
    def init_guess(self, init_guess):
        self.BK.init_guess = [-i/2 for i in init_guess]

    @property 
    def param_to_ex_ops_bra(self):
        if self.params_bra is None: return {}
        d = defaultdict(list)
        for i, j in enumerate(self.params_bra):
            if hasattr(j,'extract_variables'):
                for v in j.extract_variables():
                    d[v].append(self.bra[i])
            else:
                d[j].append(self.bra[i])
        return d
    
    
    @property
    def param_to_ex_ops_ket(self):
        d = defaultdict(list)
        for i, j in enumerate(self.params_ket):
            if hasattr(j,'extract_variables'):
                for v in j.extract_variables():
                    d[v].append(self.ket[i])
            else:
                d[j].append(self.ket[i])
        return d
    
    @property
    def param_to_ex_ops(self):
        db = self.param_to_ex_ops_bra if self.params_bra is not None else {}
        dk = self.param_to_ex_ops_ket
        db.update(dk)
        return db

    def __str__(self):
        res = ''
        if self.is_diagonal:
            res += f"{self._name} with indices: {self.ket} with variables {self.params_ket}"
        else:
            res += f"{self._name} with Bra= {self.bra} with variables {self.params_bra}\n"
            res += f"{len(self._name)*' '} with Ket= {self.ket} with variables {self.params_ket}"
        return res

    def __repr__(self):
        return self.__str__()
    
    @property
    def is_diagonal(self):
        return self.BK.is_diagonal()

    @property
    def U(self):
        'Dummy function to work with tequila Objectives'
        if self.is_diagonal:
            return self.ket 
        else:
            return [self.bra,self.ket]
    
    def count_measurements(self)->int:
        mes = 0
        if self.BK.int1e is not None:
            mes += prod(self.BK.int1e.shape)
        if self.BK.int2e is not None:
            mes += prod(self.BK.int2e.shape)
        if mes:
            return mes
        else:
            return len(self.BK.civector(params=[.0 for _ in range(self.BK.n_variables_ket)]))

    def build_operator(self,operator:Union[str,FermionOperator,QubitHamiltonian]=None)->Union[None,Callable]:
        '''
        Build the expectation value operator. 
        Even if it is accepted a QubitHamiltonian, we disencorage its use here since TCC works on fermionic states.
        It will be applied to the ci_vector and kept these results which doesn't leave the ci_vector space with READ Coeff.
        '''
        def from_string(operator:str):
            if operator.upper()=="I":
                nmo = len(self.BK.aslst)
                self.BK.hamiltonian = None
                self.BK.int1e = zeros((nmo,nmo))
                int2e = zeros((nmo,nmo,nmo,nmo))
                for i in range(nmo):
                    int2e[i,i,i,i] = 1#0.5
                self.BK.int2e = int2e
                self.BK.e_core = 0.
                self.BK.hamiltonian_lib = {}
            elif operator.upper() == "H":
                pass
            else:
                raise TequilaException(f"No operator str {operator} supported on TCC BraKet")
        
        if operator is None:
            return None
        if isinstance(operator,str):
            from_string(operator)
            return None
        elif isinstance(operator,FermionOperator):
            operator = reorder(operator=operator,order_function=up_then_down,num_modes=2*len(self.BK.aslst))
            operator = jordan_wigner(operator)
            operator.compress()
            operator = QubitHamiltonian.from_openfermion(operator)
            self.BK.e_core = 0
        elif isinstance(operator,QubitHamiltonian):
            self.BK.e_core = 0
        else:
            raise TequilaException(f"No operator {type(operator).__name__} supported")
        ci_vec = get_ci_strings(n_elec_s=self.BK.n_elec,n_qubits=2*len(self.BK.aslst),mode='fermion')
        self.BK.e_core = 0
        def callable_operator(ci_vect:List[int],operator:QubitHamiltonian,n_qubits:int, ket:List[float])->List[float]:
            hket = zeros(len(ci_vect))
            wfv = QubitWaveFunction(n_qubits=n_qubits,numbering=BitNumbering.LSB,dense=True)
            for idx,val in enumerate(ket):
                wfv._state[ci_vect[idx]] = val
            hket = wfv.apply_qubitoperator(operator)
            hket = hket.to_array(out_numbering=BitNumbering.LSB)
            return [hket[i].real for i in ci_vect]
        return lambda ket: callable_operator(ci_vect=ci_vec,operator=operator,n_qubits=2*len(self.BK.aslst), ket=ket)

def init_state_from_wavefunction(wvf:QubitWaveFunction):
    if not isinstance(wvf._state,dict):
        return init_state_from_array(wvf=wvf)
    init_state = []
    for i in wvf._state:
        vec = bin(i)[2:]
        if len(vec) < wvf.n_qubits:
            vec = '0'*(wvf.n_qubits-len(vec))+vec
        init_state.append([vec,wvf._state[i].real])#tcc automatically does this, but with an anoying message everytime
    return init_state

def init_state_from_array(wvf:QubitWaveFunction,tol=1e-6):
    '''
    Expected Initial State in UptheDown
    '''
    if isinstance(wvf._state,dict):
        return init_state_from_wavefunction(wvf)
    init_state = []
    nq = wvf.n_qubits
    nq = int(2* (ceil(nq/2))) #if HCB may be odd amount of qubits
    for i,idx in enumerate(wvf._state):
        vec = BitString.from_int(i)
        vec.nbits = nq
        vec = vec.binary
        if abs(idx) > tol:
            init_state.append([vec,idx.real]) #tcc automatically does this, but with an anoying message everytime
    return init_state

def indices_tq_to_tcc(indices:List[List[Tuple[int]]]=None):
    '''
    Expected indices like [[(0,1),(n_mo+0,n_mo+1),...],[(a,b),(c,d),...],...] (in upthendown)
    Returned [(...,n_mo+1,1,0,n_mo+0,...),(...,d,b,a,c,...)]
    '''
    if indices is None:
        return None,None,None
    if not len(indices):
        return [],[],[]
    assert isinstance(indices,(list,tuple))
    if isinstance(indices[0],Number):
        indices = [indices]
    ex_ops = []
    params = []
    param_ids = []
    for exct in indices:
        exc = []
        params.append(str(exct))
        param_ids.append(len(param_ids))
        for idx in exct:
            exc.append(idx[0])
            exc.insert(0,idx[1])
        ex_ops.append(tuple(exc))
    return ex_ops,params,param_ids

def indices_tcc_to_tq(indices:List[List[Tuple[int]]]=None)->List[List[Tuple[int]]]:
    '''
    Expected indices like [(...,n_mo+1,1,0,n_mo+0,...),(...,d,b,a,c,...)]
    Returned [[(0,1),(n_mo+0,n_mo+1),...],[(a,b),(c,d),...],...] (in upthendown)
    '''
    if indices is None:
        return None
    assert isinstance(indices,(list,tuple))
    if not len(indices):
        return []
    if isinstance(indices[0],Number):
        indices = [indices]
    ex_ops = []
    for exct in indices:
        exc = []
        for i in range(len(exct)//2):
            exc.append((exct[-i-1],exct[i]))
        ex_ops.append(exc[::-1])
    return ex_ops

def map_variables(x:list[Variable,Objective],dvariables:dict):
    if isinstance(x,Variable):
        x = x.map_variables(dvariables)
    elif isinstance(x,Objective):
        x=simulate(x,dvariables)
    return x


import numbers
from typing import List

from ..core import CircuitState, CircuitStyle
from ..quantum_chemistry import PairCorrelatorGate
from ..core import Polygon


class DoubleExcitation(PairCorrelatorGate):
    """
            i,j,k,l correspond to Spin Orbital index --> ((i,j),(k,l))
    """

    def __init__(self, i, j, k, l, angle, control=None, assume_real=True, encoding="JordanWigner", n_qubits_is_double:bool=False, *args, **kwargs):
        super().__init__(i=i, j=j, angle=angle, control=control, assume_real=assume_real, encoding=encoding, *args,**kwargs)
        self.n_qubits_is_double = n_qubits_is_double
        self.k = k
        self.l = l
        self.type = 3
        if not self.n_qubits_is_double:
            if self.i // 2 == self.k // 2 and self.j // 2 == self.l // 2:
                self.type = 0  # Actually its a UC
            elif self.i // 2 == self.k // 2 and self.j // 2 != self.l // 2:
                self.type = 1  # Origin Orbital Paired, Destination Unpaired
            elif self.i // 2 != self.k // 2 and self.j // 2 == self.l // 2:
                self.type = 2  # Origin Orbital Unpaired, Destination Paired
            else:
                self.type = 3  # Completely Unpaired


    def used_wires(self) -> List[int]:
        used = list(range(min(self.i, self.j, self.k, self.l) // (1+(not self.n_qubits_is_double)), max(self.i, self.j, self.k, self.l) // (1+(not self.n_qubits_is_double)) + 1))

        if self.control is not None:
            used.append(self.control // (1+(not self.n_qubits_is_double)))
        return used

    def construct_circuit(self, *args, **kwargs):
        if not self.type:
            return PairCorrelatorGate(i=self.i // 2, j=self.j // 2, assume_real=self.assume_real, angle=self.angle,encoding=self.encoding).construct_circuit(*args, **kwargs)
        else:
            return self._dummy.make_excitation_gate(indices=((self.i, self.j), (self.k, self.l)),assume_real=self.assume_real, angle=self.angle,control=self.control, *args, **kwargs)

    def _render(self, state: CircuitState, style: CircuitStyle) -> str:
        spin = {0: Polygon(-3), 1: Polygon(3)}

        # styling
        tcol = style.secondary.text_color
        gcol = style.secondary.gate_color

        # angle
        gcol, tcol = self.render_angle(style, gcol, tcol)
        if 'double' in style.personalized:
            gcol = style.personalized['double'].gate_color
            tcol = style.personalized['double'].text_color
        # TODO: param is not used again in the original??
        param = self.angle
        if isinstance(param, numbers.Number):
            param = "{:1.2f}".format(param)

        # render
        result = ""
        if not self.type:
            shape = Polygon(6)
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i // 2,shape=shape,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j // 2,shape=shape,gcol=gcol,tcol="{" + tcol.name + "}",op="")
        elif self.type == 1:
            shape1 = Polygon(6)
            shape2 = spin[self.j % 2]
            shape3 = spin[self.l % 2]
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i // 2,shape=shape1,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j // 2,shape=shape2,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.l // 2,shape=shape3,gcol=gcol,tcol="{" + tcol.name + "}",op="")
        elif self.type == 2:
            shape3 = Polygon(6)
            shape1 = spin[self.i % 2]
            shape2 = spin[self.k % 2]
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i // 2,shape=shape1,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.k // 2,shape=shape2,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j // 2,shape=shape3,gcol=gcol,tcol="{" + tcol.name + "}",op="")
        else:
            shape1 = spin[self.i % 2]
            shape2 = spin[self.j % 2]
            shape3 = spin[self.k % 2]
            shape4 = spin[self.l % 2]
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i // (1+(not self.n_qubits_is_double)),shape=shape1,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j // (1+(not self.n_qubits_is_double)),shape=shape2,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.k // (1+(not self.n_qubits_is_double)),shape=shape3,gcol=gcol,tcol="{" + tcol.name + "}",op="")
            result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.l // (1+(not self.n_qubits_is_double)),shape=shape4,gcol=gcol,tcol="{" + tcol.name + "}",op="")
        return result


import numpy as np
import tequila as tq
from sunrise import OrbitalRotation as OR


def test_Orbital_Rotation():
    idx = [0, 1]
    a = tq.QTensor(shape=(2, 2))
    a[0, 0] = tq.Variable("a00")
    a[0, 1] = tq.Variable('a01')
    a[1, 0] = tq.Variable('a10')
    a[1, 1] = tq.Variable('a11')
    RA = OR(orbitals=idx, matrix=a)
    variables = {'a00': 0, 'a01': 1, 'a10': 1, 'a11': 0}
    wfn_a = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string('1*|0011>')
    assert wfn_a.isclose(tq.simulate(tq.gates.X([0,1]) + RA, variables=variables))

    jdx = [1, 2]
    b = tq.QTensor(shape=(2, 2))
    b[0, 0] = tq.Variable("b00")
    b[0, 1] = tq.Variable("b00")
    b[1, 0] = tq.Variable("b00")
    b[1, 1] = tq.Variable("b00")
    variables.update({"b00": 1, "b01": -1, "b10": 1, "b11": 1})
    b = (1 / (np.sqrt(2))) * b
    RB = OR(orbitals=jdx,matrix=b)
    wfn_b = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string('0.5*|001100> - 0.5*|000110> + 0.5*|001001> + 0.5|000011>')
    assert wfn_b.isclose(tq.simulate(tq.gates.X([2,3]) + RB, variables=variables))

    RC = RA + RB
    RD = RB + RA
    wfn_c = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string('- 0.5*|001100> + 0.5*|000110> - 0.5*|001001> - 0.5*|000011>')
    wfn_d = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string('-1*|110000>')
    wfn_e = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string(' 1*|001100>')
    wfn_f = tq.wavefunction.qubit_wavefunction.QubitWaveFunction.from_string(' 0.5*|110000> - 0.5*|010010> + 0.5*|100001> + 0.5*|000011>')

    assert wfn_c.isclose(tq.simulate(objective=tq.gates.X([0,1])+RC, variables=variables))
    assert wfn_d.isclose(tq.simulate(objective=tq.gates.X([2,3])+RC, variables=variables))
    assert wfn_e.isclose(tq.simulate(objective=tq.gates.X([0,1])+RD, variables=variables))
    assert wfn_f.isclose(tq.simulate(objective=tq.gates.X([2,3])+RD, variables=variables))



import tequila as tq
import numpy as np

class input_state:
    """
    Wrapper class to hold the input state to measure the molecule Hamiltonian
    The input state can be a quantum circuit or a wavefunction
    """
    circuit = None
    variables = None
    wavefunction = None

    def __init__(self, circuit=None, variables=None, wavefunction=None):
        if circuit is None and wavefunction is None:
            raise ValueError("Either a circuit or a wavefunction must be provided")
        if circuit is not None and wavefunction is not None:
            raise ValueError("Only one of circuit or wavefunction must be provided")
        
        self.circuit = circuit
        self.variables = variables
        self.wavefunction = wavefunction

    def get_circuit(self):
        """
        Get the circuit from the wavefunction or the circuit
        """
        if self.wavefunction is not None and self.circuit is None:
            self.circuit = tq.gates.Rz(angle=0, target=range(self.wavefunction.n_qubits)) # dummy empty circuit
        return self.circuit
        
    def get_wavefunction(self, variables=None):
        """
        Get the wavefunction from the circuit or the wavefunction
        """
        if self.circuit is not None and self.wavefunction is None:
            if variables is None:
                variables = self.variables
            self.wavefunction = tq.compile(self.circuit)(variables)
        return self.wavefunction

def gates_to_orb_rot(rotation_circuit:tq.QCircuit, dim:int, isReordered:bool=False,core:int=0)->np.array:
    """
    rotation_circuit: a circuit with ONLY UR gates
    dim: the dimension of the spatial orbital space, i.e., the number of orbitals
    core: number of core orbitals, already counted on dim. This is bcs these core orbitals are not included on the UR qubit count
    Output: a matrix that rotates the orbital coefficients
    """

    # Extract parameters and indices from gates
    params = []
    indices = []
    for gate in reversed(rotation_circuit.gates):
        params.append(gate.parameter)
        g = []
        for idx in gate.indices:
            g.append(tuple([i+core+core*(not isReordered) for i in idx]))
        indices.append(g)

    # Select only odd values
    # Because UR gates rotate molecular orbitals 
    params = params[1::2]
    indices = indices[1::2]

    # Refactor as a list
    tmp = []
    for i in indices:
        if isReordered:
            tmp.append((int(i[0][0]),int(i[0][1])))
        else:
            tmp.append((int(i[0][0]/2),int(i[0][1]/2)))
    indices = tmp

    # Create a matrix for each UR gate
    rot_matrices = []
    for k,index in enumerate(indices):
        tmp = np.eye(dim) # dim is the dimension of orbital coefficients
        if isinstance(params[k], tq.Objective):
            params[k] = params[k]()
        tmp[index[0]][index[0]] = np.cos(params[k]/2)
        tmp[index[0]][index[1]] = np.sin(params[k]/2)
        tmp[index[1]][index[0]] = -np.sin(params[k]/2)
        tmp[index[1]][index[1]] = np.cos(params[k]/2)
        rot_matrices.append(tmp)

    # Multiply all matrices
    tmp = rot_matrices[0]
    for matrix in rot_matrices[1:]:
        tmp = np.dot(tmp,matrix)
    transformation_matrix = tmp

    return transformation_matrix

def fold_rotators(mol, UR, *args, **kwargs):
    """
    Rotate the one- and two-body integrals of a molecule using the UR gates
    """
    # Get the transformation matrix from the circuit
    UR_matrix = gates_to_orb_rot(UR, mol.n_orbitals, *args, **kwargs)

    # Rotate one- and two-body part
    c,h,g = mol.get_integrals()
    g = g.elems

    th = np.einsum("ix, jx -> ij",  h, UR_matrix, optimize='greedy')
    th = np.einsum("xj, ix -> ij", th, UR_matrix, optimize='greedy')
    # same as th = (UR_matrix.dot(h)).dot(UR_matrix.T)

    tg = np.einsum("ijkx, lx -> ijkl",  g, UR_matrix, optimize='greedy')
    tg = np.einsum("ijxl, kx -> ijkl", tg, UR_matrix, optimize='greedy')
    tg = np.einsum("ixkl, jx -> ijkl", tg, UR_matrix, optimize='greedy')
    tg = np.einsum("xjkl, ix -> ijkl", tg, UR_matrix, optimize='greedy')

    tmol = tq.Molecule(parameters=mol.parameters, transformation=mol.transformation, n_electrons=mol.n_electrons,
                       nuclear_repulsion=c, one_body_integrals=th, two_body_integrals=tg, ordering='openfermion')

    return tmol

def get_hcb_part(mol):
    """
    Extract the hcb part and the residual part of a molecule
    """
    c,h,g = mol.get_integrals()

    # Get the HCB part of the one-body integral
    hcb_h = np.zeros(shape=(h.shape[0], h.shape[1]))
    res_h = np.zeros(shape=(h.shape[0], h.shape[1]))
    for i in range(h.shape[0]):
        for j in range(h.shape[1]):
            if i==j:
                hcb_h[i][j] = h[i][j] # ii
            else:
                res_h[i][j] = h[i][j]

    # Get the HCB part of the two-body integral
    hcb_g = np.zeros(shape=(g.shape[0], g.shape[1], g.shape[2], g.shape[3]))
    res_g = np.zeros(shape=(g.shape[0], g.shape[1], g.shape[2], g.shape[3]))
    for i in range(g.shape[0]):
        for j in range(g.shape[1]):
            for k in range(g.shape[2]):
                for l in range(g.shape[3]):
                    if (i==j and k==l) or (i==l and j==k) or (i==k and j==l): # iikk or ikki or ijij
                        hcb_g[i][j][k][l] = g.elems[i][j][k][l]
                    else:
                        res_g[i][j][k][l] = g.elems[i][j][k][l]

    # Create the HCB and residual molecule
    hcb_mol = tq.Molecule(parameters=mol.parameters, transformation=mol.transformation, n_electrons=mol.n_electrons,
                          nuclear_repulsion=c, one_body_integrals=hcb_h, two_body_integrals=hcb_g, ordering='openfermion')
    res_mol = tq.Molecule(parameters=mol.parameters, transformation=mol.transformation, n_electrons=mol.n_electrons,
                          nuclear_repulsion=0, one_body_integrals=res_h, two_body_integrals=res_g, ordering='openfermion')

    return hcb_mol, res_mol

def rotate_and_hcb(molecule, rotators, circuit=None, variables=None, initial_state=None, target=None, silent=True, *args, **kwargs):
    """
    Rotate the molecule using the UR gates and extract the hcb part and the residual part of the transformed molecule.
    Iterate for all the elements of the rotators list.

    molecule: the molecule that holds the Hamiltonian to be processed
    rotators: the UR gates to be applied

    circuit: the quantum circuit to be measured
    variables: the parameters of the circuit
    initial_state: if you want to use a wavefunction without a quantum circuit

    approx: the initial approximation of the energy
    target: the target energy

    Return a list of HCB molecules and the last residual molecule
    """

    # Initialize the input state in a wrapper class
    state = input_state(circuit=circuit, variables=variables, wavefunction=initial_state)
    if initial_state is not None and circuit is None:
        circuit = state.get_circuit()
        initial_state = state.get_wavefunction()
    elif circuit is None and initial_state is None:
        raise ValueError("Either a circuit or a wavefunction must be provided")

    # General procedure, loop over the list of rotators
    hcb_mols = []
    approx = 0.0
    old_UR = tq.QCircuit()
    for i,UR in enumerate(rotators):
        # Rotate the two body part
        tmol = fold_rotators(molecule, old_UR.dagger()+UR, *args, **kwargs)
        tcircuit = circuit + UR

        # Extract the hcb part and the residual part of the transformed molecule
        hcb_mol, res_mol = get_hcb_part(tmol)

        if not silent:
            # Compute expectation values and error
            EX = tq.ExpectationValue(U=tcircuit, H=hcb_mol.make_hamiltonian())
            incr = tq.simulate(EX, variables=variables, initial_state=initial_state, *args, **kwargs)
            EY = tq.ExpectationValue(U=tcircuit, H=res_mol.make_hamiltonian())
            rest = tq.simulate(EY, variables=variables, initial_state=initial_state, *args, **kwargs)
            approx += incr

            print(f"Iteration {i+1}/{len(rotators)}:")

            print(f"Approximation:         {approx}")
            print(f"Increment:             {incr}")
            print(f"Error in new basis:    {target-approx}")
            # print(f"test:                  {target-approx-rest}")

            M_tot = compute_num_meas(circuit=circuit, variables=variables, initial_state=initial_state, is_hcb=True, hcb_mol=hcb_mol)
            print(f"Number of measurements: {M_tot:e}\n")

        # Update values
        hcb_mols.append(hcb_mol)
        molecule = res_mol
        old_UR = UR

    return hcb_mols, res_mol

def compute_num_meas(circuit=None, variables=None, initial_state=None, is_hcb=False, hcb_mol=None, meas_groups=None, transformations=None, eps=1e-3):
    """
    Compute the number of measurements needed to estimate the expectation value of a Hamiltonian.
    
    circuit: the quantum circuit to be measured
    variables: the parameters of the circuit
    initial_state: if you want to use a wavefunction without a quantum circuit

    is_hcb: if True, we measure the HCB Hamiltonian exploiting the three measurement groups
    hcb_mol: the molecule that holds the HCB Hamiltonian
    meas_groups: the measurement groups, in case the operator is not the HCB Hamiltonian
    transformations: the transformations to be applied to the circuit, in case the operator is not the HCB Hamiltonian
    eps: the error tolerance
    """
    
    # If hcb then split in the three measurement groups
    if is_hcb:
        H1 = tq.QubitHamiltonian()
        H2 = tq.QubitHamiltonian()
        H3 = tq.QubitHamiltonian()
        for p in hcb_mol.make_hamiltonian().paulistrings:
            q = p.naked().qubits
            if p.is_all_z():
                H1 += tq.QubitHamiltonian().from_paulistrings(p)
            else:
                if (p.naked()[q[0]] == "X" and p.naked()[q[1]] == "X") or (p.naked()[q[0]] == "Y" and p.naked()[q[1]] == "Y"):
                    H2 += tq.QubitHamiltonian().from_paulistrings(p)
                else:
                    H3 += tq.QubitHamiltonian().from_paulistrings(p)
        meas_groups = [H1, H2, H3]

    # Compute group number of measurements
    M_tot = 0
    for i, group in enumerate(meas_groups):

        if is_hcb:
            # Diagonalize each of the three hcb measurement groups
            fc_groups_and_unitaries, _ = tq.grouping.compile_groups.compile_commuting_parts(group, unitary_circuit='improved')
            group = fc_groups_and_unitaries[0][0]
            transformation = fc_groups_and_unitaries[0][1]
        else:
            transformation = transformations[i]

        # Compute individual number of measurements
        M_group = 1
        for op in group.paulistrings:
            # Compute individual Pauli number of measurements
            pauli_exp = tq.ExpectationValue(circuit+transformation, tq.QubitHamiltonian().from_paulistrings(op.naked()))
            var = 1.0 - tq.simulate(pauli_exp, variables=variables, initial_state=initial_state)**2
            M_l = ((abs(op.coeff) * np.sqrt(var)) / eps) ** 2

            # Pick the largest number of measurements
            if M_l > M_group:
                M_group = M_l
        
        M_tot += M_group

    return M_tot